from .colors import Colors
from .utils import UINT256_MAX, decode_printable_with_color, pad_even


class Stack:
    stack: list[int]
    ignore_stack_underflow: bool

    updated_indices_for_colorize: list[int]

    def __init__(self, ignore_stack_underflow: bool = False) -> None:
        self.stack = []
        self.ignore_stack_underflow = ignore_stack_underflow

        self.updated_indices_for_colorize = []

    def push(self, x: int) -> None:
        assert 0 <= x <= UINT256_MAX
        self.stack.append(x)

        self.updated_indices_for_colorize = [len(self.stack) - 1]

    def extend(self, x: list[int]) -> None:
        assert 0 <= max(x) <= UINT256_MAX
        self.stack.extend(x)

        self.updated_indices_for_colorize = [len(self.stack) - 1 - i for i in range(len(x))]

    def pop(self) -> int:
        if len(self.stack) == 0:
            if self.ignore_stack_underflow:
                return 0
            else:
                raise Exception("Stack underflow")
        return self.stack.pop()

    def __getitem__(self, i: int) -> int:
        if i >= len(self.stack):
            if self.ignore_stack_underflow:
                return 0
            else:
                raise Exception("Stack underflow")
        return self.stack[i]

    def clear(self) -> None:
        self.stack = []

    def to_string(self) -> str:
        ret = ""
        n = len(self.stack)
        for i in range(n):
            if i != 0:
                ret += ", "
            if isinstance(self.stack[n - 1 - i], int):
                x = pad_even(hex(self.stack[n - 1 - i]))
            else:
                x = str(self.stack[n - 1 - i])
            if n - 1 - i in self.updated_indices_for_colorize:
                x = Colors.GREEN + x + Colors.ENDC
            ret += x
        ret = "[" + ret + "]"
        return ret

    def to_string_with_decode(self) -> str:
        ret = ""
        n = len(self.stack)
        for i in range(n):
            if i != 0:
                ret += ", "
            if isinstance(self.stack[n - 1 - i], int):
                x = decode_printable_with_color(pad_even(hex(self.stack[n - 1 - i]))[2:])
            else:
                x = str(self.stack[n - 1 - i])
            if n - 1 - i in self.updated_indices_for_colorize:
                x = Colors.GREEN + x + Colors.ENDC
            ret += x
        ret = "[" + ret + "]"
        return ret

    def __len__(self) -> int:
        return len(self.stack)



from enum import Enum

from .context import Context
from .disassemble import disassemble
from .opcodes import OPCODES
from .utils import UINT256_MAX, pad


class ControlType(Enum):
    BEFORE_JUMPDEST = 0
    JUMP = 1
    JUMPI = 2
    END = 3
    BAD = -1


def disassemble_mermaid(context: Context, entrypoint: int = 0x00, max_steps: int = UINT256_MAX) -> None:
    """
    ブロックの開始
    - 0x00
    - JUMPDEST
    - JUMPIの一つ後
    ブロックの終了
    - JUMP
    - JUMPDESTの一つ前
    - REVERT
    - INVALID
    - SELFDESTRUCT
    - STOP
    - RETURN
    """

    LOCATION_PAD_N = len(hex(len(context.bytecode))[2:])

    start_addresses = [0x00]

    for i in range(len(context.bytecode)):
        value = context.bytecode[i]
        if value not in OPCODES:
            continue
        mnemonic = OPCODES[value][0]
        if mnemonic == "JUMPDEST":
            # JUMPIの一つ後で追加済みならば追加しない
            if start_addresses[-1] != i:
                start_addresses.append(i)
        elif mnemonic == "JUMPI":
            start_addresses.append(i + 1)

    def disassemble_block(
        start_address: int,
    ) -> tuple[bool, int, ControlType, list[str]]:
        """
        return is_valid_block, end_address, control_type, instructions
        """

        pc = start_address
        instructions = []
        while pc < len(context.bytecode):
            next_pc = pc + 1
            value = context.bytecode[pc]

            if value in OPCODES:
                (
                    mnemonic,
                    stack_input_count,
                    _stack_output_count,
                    _base_gas,
                    _description,
                    stack_input_names,
                ) = OPCODES[value]
            else:
                instructions.append(f"{pad(hex(pc), LOCATION_PAD_N)}: 0x (?)")
                return False, pc, ControlType.BAD, instructions

            if mnemonic.startswith("PUSH"):
                mnemonic_num = int(mnemonic[4:])
                next_pc = pc + 1 + mnemonic_num
            elif mnemonic.startswith("DUP"):
                mnemonic_num = int(mnemonic[3:])
            elif mnemonic.startswith("SWAP"):
                mnemonic_num = int(mnemonic[4:])
            elif mnemonic.startswith("LOG"):
                mnemonic_num = int(mnemonic[3:])

            if mnemonic.startswith("PUSH"):
                instructions.append(
                    f"{pad(hex(pc), LOCATION_PAD_N)}: {mnemonic}  0x{context.bytecode[pc+1:pc+1+mnemonic_num].hex()}"
                )
            else:
                instructions.append(f"{pad(hex(pc), LOCATION_PAD_N)}: {mnemonic}")

            if pc != start_address and mnemonic == "JUMPDEST":
                return True, pc - 1, ControlType.BEFORE_JUMPDEST, instructions
            elif mnemonic == "JUMP":
                return True, pc, ControlType.JUMP, instructions
            elif mnemonic == "JUMPI":
                return True, pc, ControlType.JUMPI, instructions
            elif mnemonic in ["REVERT", "INVALID", "SELFDESTRUCT", "STOP", "RETURN"]:
                return True, pc, ControlType.END, instructions

            pc = next_pc

        assert False

    graph = ""
    for start_address in start_addresses:
        is_valid_block, end_address, control_type, instructions = disassemble_block(start_address)
        if not is_valid_block:
            continue
        block = "\\n".join(instructions)
        max_steps = len(instructions)
        error = False
        try:
            last_jump_to_address = disassemble(
                context=context,
                trace=True,
                entrypoint=start_address,
                max_steps=max_steps,
                decode_stack=False,
                ignore_stack_underflow=True,
                silent=True,
            ).last_jump_to_address
        except Exception:
            error = True

        block_id = pad(hex(start_address), LOCATION_PAD_N) if start_address != entrypoint else "START"
        if error:
            graph += f"{block_id}({block}) --> ERROR\n"
            continue
        match control_type:
            case ControlType.BEFORE_JUMPDEST:
                next_block_id = pad(hex(end_address), LOCATION_PAD_N)
                graph += f"{block_id}({block}) --> {next_block_id}\n"
            case ControlType.JUMP:
                assert isinstance(last_jump_to_address, int)
                next_block_id = pad(hex(last_jump_to_address), LOCATION_PAD_N)
                graph += f"{block_id}({block}) --jump--> {next_block_id}\n"
            case ControlType.JUMPI:
                assert isinstance(last_jump_to_address, int)
                next_block_id = pad(hex(last_jump_to_address), LOCATION_PAD_N)
                graph += f"{block_id}({block}) --jump--> {next_block_id}\n"
                next_block_id = pad(hex(end_address + 1), LOCATION_PAD_N)
                graph += f"{block_id} --> {next_block_id}\n"
            case ControlType.END:
                next_block_id = "END"
                graph += f"{block_id}({block}) --> {next_block_id}\n"

    print(
        """<html lang="en">
    <head>
        <meta charset="utf-8" />
    </head>
    <body>"""
        + f"""<pre class="mermaid">

flowchart TB 
{graph}
    </pre>"""
        + """<script type="module">
        import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@10/dist/mermaid.esm.min.mjs';
        mermaid.initialize({ startOnLoad: true });
    </script>
    <style>
    .mermaid .node .label {
        text-align: left !important;
    }
    </style>
    </body>
    </html>
    """
    )



import argparse
import json
import os
import sys
import tomllib

from .assemble import assemble
from .colors import Colors
from .context import Context
from .disassemble import MemoryDisplay, disassemble
from .disassemble_mermaid import disassemble_mermaid
from .disassemble_symbolic import disassemble_symbolic
from .find_gadgets import find_gadgets
from .utils import UINT256_MAX


class CustomHelpFormatter(argparse.HelpFormatter):
    def __init__(
        self,
        prog: str,
        indent_increment: int = 2,
        max_help_position: int = 53,
        width: int | None = None,
    ) -> None:
        super().__init__(prog, indent_increment, max_help_position, width)

    def _format_action_invocation(self, action: argparse.Action) -> str:
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            (metavar,) = self._metavar_formatter(action, default)(1)
            return Colors.BOLD + metavar + Colors.ENDC
        else:
            if action.nargs != 0:
                return (
                    Colors.BOLD
                    + ", ".join(action.option_strings)
                    + Colors.ENDC
                    + " "
                    + self._format_args(action, action.dest.upper())
                )
            else:
                return Colors.BOLD + ", ".join(action.option_strings) + Colors.ENDC

    def _get_help_string(self, action: argparse.Action) -> str:
        help = action.help
        if help is None:
            help = ""

        if "%(default)" not in help:
            if action.default is not argparse.SUPPRESS:
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    if action.default == str(UINT256_MAX):
                        help += " (default: (1 << 256) - 1)"
                    else:
                        help += " (default: %(default)s)"
        return help


def main() -> None:
    parser = argparse.ArgumentParser(
        description="EVM Reversing Tools",
        formatter_class=CustomHelpFormatter,
    )
    subparsers = parser.add_subparsers()

    parser_disassemble = subparsers.add_parser(
        "disassemble",
        aliases=["disas"],
        help="Disassemble the given bytecode",
        formatter_class=CustomHelpFormatter,
    )
    parser_disassemble.set_defaults(handler=command_disassemble)

    parser_trace = subparsers.add_parser(
        "trace",
        help="Trace execution of the given bytecode",
        formatter_class=CustomHelpFormatter,
    )
    parser_trace.set_defaults(handler=command_trace)

    parser_symbolic_trace = subparsers.add_parser(
        "symbolic-trace",
        aliases=["symbolic"],
        help="Trace execution of the given bytecode symbolically",
        formatter_class=CustomHelpFormatter,
    )
    parser_symbolic_trace.set_defaults(handler=command_symbolic_trace)

    parser_mermaid = subparsers.add_parser(
        "mermaid",
        help="Generate the mermaid diagram for the given bytecode",
        formatter_class=CustomHelpFormatter,
    )
    parser_mermaid.set_defaults(handler=command_mermaid)

    parser_gadget = subparsers.add_parser(
        "gadget",
        help="Find JOP gadgets in the given bytecode",
        formatter_class=CustomHelpFormatter,
    )
    parser_gadget.set_defaults(handler=command_gadget)

    parser_assemble = subparsers.add_parser(
        "assemble",
        help="Assemble the given mnemonics to the bytecode",
        formatter_class=CustomHelpFormatter,
    )
    parser_assemble.set_defaults(handler=command_assemble)

    def add_common_arguments_for_data(
        parser: argparse.ArgumentParser,
    ) -> argparse._ArgumentGroup:
        group = parser.add_argument_group("data options")

        group.add_argument("-b", "--bytecode", help="Raw bytecode as hex string")
        group.add_argument("-f", "--filename", help="TOML or bytecode file")

        group.add_argument(
            "--rpc-url",
            type=str,
            default=os.getenv("EREVER_RPC_URL"),
            help="Fetch state from a remote endpoint instead of empty state",
        )
        group.add_argument("-c", "--contract-address", help="Requires --rpc-url")
        group.add_argument("--tx", help="Requires --rpc-url")

        group.add_argument(
            "--entrypoint",
            type=str,
            default="0",
            help="PC of entrypoint for disassembly/tracing",
        )
        group.add_argument(
            "--max-steps",
            type=str,
            default=str(UINT256_MAX),
            help="Maximum number of steps to disassemble/trace",
        )
        return group

    def add_common_arguments_for_display(
        parser: argparse.ArgumentParser,
    ) -> argparse._ArgumentGroup:
        group = parser.add_argument_group("display options")
        group.add_argument("--show-opcodes", action="store_true", default=False, help="Show opcodes")
        group.add_argument("--hide-pc", action="store_true", default=False, help="Hide PC")
        return group

    def add_common_arguments_for_constructing_context(
        parser: argparse.ArgumentParser,
    ) -> argparse._ArgumentGroup:
        group = parser.add_argument_group("context options")

        group.add_argument(
            "--address",
            type=str,
            default=str(Context.DEFAULT_ADDRESS),
            help="Address of the contract",
        )
        group.add_argument(
            "--origin",
            type=str,
            default=str(Context.DEFAULT_ORIGIN),
            help="Origin of the transaction",
        )
        group.add_argument(
            "--caller",
            type=str,
            default=str(Context.DEFAULT_CALLER),
            help="Caller of the transaction",
        )
        group.add_argument(
            "--callvalue",
            type=str,
            default=str(Context.DEFAULT_CALLVALUE),
            help="Call value of the transaction",
        )
        group.add_argument(
            "--calldata",
            type=str,
            default=Context.DEFAULT_CALLDATA.decode(),
            help="Call data of the transaction",
        )
        group.add_argument(
            "--gasprice",
            type=str,
            default=str(Context.DEFAULT_GASPRICE),
            help="Gas price of the transaction",
        )
        group.add_argument(
            "--coinbase",
            type=str,
            default=str(Context.DEFAULT_COINBASE),
            help="Coinbase of the block",
        )
        group.add_argument(
            "--timestamp",
            type=str,
            default=str(Context.DEFAULT_TIMESTAMP),
            help="Timestamp of the block",
        )
        group.add_argument(
            "--number",
            type=str,
            default=str(Context.DEFAULT_NUMBER),
            help="Number of the block",
        )
        group.add_argument(
            "--difficulty",
            type=str,
            default=str(Context.DEFAULT_DIFFICULTY),
            help="Difficulty of the block",
        )
        group.add_argument(
            "--gaslimit",
            type=str,
            default=str(Context.DEFAULT_GASLIMIT),
            help="Gas limit of the block",
        )
        group.add_argument(
            "--chainid",
            type=str,
            default=str(Context.DEFAULT_CHAINID),
            help="Chain ID of the block",
        )
        group.add_argument(
            "--selfbalance",
            type=str,
            default=str(Context.DEFAULT_SELFBALANCE),
            help="Balance of the contract",
        )
        group.add_argument(
            "--basefee",
            type=str,
            default=str(Context.DEFAULT_BASEFEE),
            help="Base fee of the block",
        )
        group.add_argument(
            "--gas",
            type=str,
            default=str(Context.DEFAULT_GAS),
            help="Gas of the transaction",
        )
        return group

    add_common_arguments_for_data(parser_disassemble)
    add_common_arguments_for_data(parser_trace)
    add_common_arguments_for_data(parser_symbolic_trace)
    add_common_arguments_for_data(parser_mermaid)
    add_common_arguments_for_data(parser_gadget)
    add_common_arguments_for_display(parser_disassemble)
    trace_display_group = add_common_arguments_for_display(parser_trace)
    symbolic_trace_display_group = add_common_arguments_for_display(parser_symbolic_trace)
    add_common_arguments_for_display(parser_mermaid)
    add_common_arguments_for_display(parser_gadget)
    add_common_arguments_for_constructing_context(parser_trace)
    add_common_arguments_for_constructing_context(parser_symbolic_trace)
    add_common_arguments_for_constructing_context(parser_mermaid)
    add_common_arguments_for_constructing_context(parser_gadget)

    trace_display_group.add_argument("--decode-stack", action="store_true", default=False, help="Decode stack items")
    trace_display_group.add_argument(
        "--invocation-only",
        action="store_true",
        default=False,
        help="Display only invocation",
    )
    trace_display_group.add_argument("--output-json", action="store_true", default=False, help="Output trace as JSON")
    trace_display_group.add_argument(
        "--return-trace-logs",
        action="store_true",
        default=False,
        help="Return trace logs",
    )
    trace_display_group.add_argument("--silent", action="store_true", default=False, help="Don't print anything")
    trace_display_group.add_argument(
        "--memory-display",
        type=MemoryDisplay,
        choices=list(MemoryDisplay),
        default=MemoryDisplay.ONCHANGE,
        help="Specify when to display memory",
    )
    trace_display_group.add_argument(
        "--memory-range",
        nargs=2,
        action="append",
        type=str,
        metavar=("START", "END"),
        help="Specify memory ranges [START, END). Not applicable to subcontexts. Use multiple times for multiple ranges.",
    )
    symbolic_trace_display_group.add_argument(
        "--show-symbolic-stack",
        action="store_true",
        default=False,
        help="Show symbolic stack items",
    )
    symbolic_trace_display_group.add_argument(
        "--hide-instructions-with-no-stack-output",
        action="store_true",
        default=False,
        help="Hide instructions which do not output to the stack",
    )
    parser_assemble.add_argument("mnemonics", metavar="MNEMONICS", type=str, help="Mnemonics to assemble")

    args = parser.parse_args()

    if not hasattr(args, "handler"):
        parser.print_help()
        exit(1)

    if args.handler == command_assemble:
        args.handler(args)
    else:
        args.rpc_url = args.rpc_url

        args.entrypoint = parse_arg_param_to_int(args.entrypoint)
        args.max_steps = parse_arg_param_to_int(args.max_steps)

        args.calldata = args.calldata if hasattr(args, "calldata") else ""
        args.address = parse_arg_param_to_int(args.address) if hasattr(args, "address") else None
        args.balance = parse_arg_param_to_int(args.balance) if hasattr(args, "balance") else None
        args.origin = parse_arg_param_to_int(args.origin) if hasattr(args, "origin") else None
        args.caller = parse_arg_param_to_int(args.caller) if hasattr(args, "caller") else None
        args.callvalue = parse_arg_param_to_int(args.callvalue) if hasattr(args, "callvalue") else None
        args.gasprice = parse_arg_param_to_int(args.gasprice) if hasattr(args, "gasprice") else None
        args.coinbase = parse_arg_param_to_int(args.coinbase) if hasattr(args, "coinbase") else None
        args.timestamp = parse_arg_param_to_int(args.timestamp) if hasattr(args, "timestamp") else None
        args.number = parse_arg_param_to_int(args.number) if hasattr(args, "number") else None
        args.difficulty = parse_arg_param_to_int(args.difficulty) if hasattr(args, "difficulty") else None
        args.gaslimit = parse_arg_param_to_int(args.gaslimit) if hasattr(args, "gaslimit") else None
        args.chainid = parse_arg_param_to_int(args.chainid) if hasattr(args, "chainid") else None
        args.selfbalance = parse_arg_param_to_int(args.selfbalance) if hasattr(args, "selfbalance") else None
        args.basefee = parse_arg_param_to_int(args.basefee) if hasattr(args, "basefee") else None
        args.gas = parse_arg_param_to_int(args.gas) if hasattr(args, "gas") else None
        args.memory_range = (
            [tuple([parse_arg_param_to_int(x) for x in y]) for y in args.memory_range]
            if hasattr(args, "memory_range") and args.memory_range is not None
            else None
        )

        if args.bytecode:
            context = Context.from_arg_params_with_bytecode(args, args.bytecode)
        elif args.filename:
            if args.filename.split(".")[-1] == "toml":
                parsed_toml = tomllib.load(open(args.filename, "rb"))
                if "bytecode" in parsed_toml:
                    parsed_toml["bytecode"] = bytes.fromhex(parsed_toml["bytecode"].replace("0x", "").replace(" ", ""))
                if "calldata" in parsed_toml:
                    parsed_toml["calldata"] = bytes.fromhex(parsed_toml["calldata"].replace("0x", "").replace(" ", ""))
                if "state" in parsed_toml:
                    parsed_toml["state_dict"] = parsed_toml["state"]
                    del parsed_toml["state"]
                context = Context(**parsed_toml)
            else:
                bytecode = open(args.filename).read()
                context = Context.from_arg_params_with_bytecode(args, bytecode)
        elif args.tx:
            context = Context.from_tx_hash(args)
        elif args.contract_address:
            context = Context.from_contract_address(args)
        else:
            parser.print_help(sys.stderr)
            exit(1)

        args.handler(args, context)


def command_disassemble(args: argparse.Namespace, context: Context) -> None:
    disassemble(
        context,
        trace=False,
        entrypoint=args.entrypoint,
        max_steps=args.max_steps,
        hide_pc=args.hide_pc,
        show_opcodes=args.show_opcodes,
    )

    # TODO: enable
    # inspect_cbor(context.bytecode)


def command_trace(args: argparse.Namespace, context: Context) -> None:
    # tx gas
    context.gas -= 21000

    # tx data gas
    for byte in context.calldata:
        if byte == 0:
            context.gas -= 4
        else:
            context.gas -= 16

    result = disassemble(
        context,
        trace=True,
        entrypoint=args.entrypoint,
        max_steps=args.max_steps,
        decode_stack=args.decode_stack,
        hide_pc=args.hide_pc,
        show_opcodes=args.show_opcodes,
        memory_display=args.memory_display,
        invocation_only=args.invocation_only,
        silent=args.silent,
        return_trace_logs=args.return_trace_logs,
        memory_range=args.memory_range,
    )
    if args.output_json:
        print(json.dumps(result.to_dict()))


def command_symbolic_trace(args: argparse.Namespace, context: Context) -> None:
    disassemble_symbolic(
        context,
        entrypoint=args.entrypoint,
        show_symbolic_stack=args.show_symbolic_stack,
        max_steps=args.max_steps,
        hide_pc=args.hide_pc,
        hide_instructions_with_no_stack_output=args.hide_instructions_with_no_stack_output,
        show_opcodes=args.show_opcodes,
    )


def command_mermaid(args: argparse.Namespace, context: Context) -> None:
    disassemble_mermaid(context, args.entrypoint, args.max_steps)


def command_gadget(args: argparse.Namespace, context: Context) -> None:
    find_gadgets(context, args.max_steps)


def command_assemble(args: argparse.Namespace) -> None:
    bytecode = assemble(args.mnemonics)
    print(bytecode.hex())


def parse_arg_param_to_int(param: str) -> int:
    if param.startswith("0x"):
        return int(param, 16)
    else:
        return int(param)


if __name__ == "__main__":
    main()



from .assemble import assemble
from .context import Context
from .disassemble import disassemble
from .disassemble_mermaid import disassemble_mermaid
from .disassemble_symbolic import disassemble_symbolic
from .find_gadgets import find_gadgets

__all__ = [
    "disassemble",
    "disassemble_mermaid",
    "disassemble_symbolic",
    "find_gadgets",
    "assemble",
    "Context",
]



from .opcodes import OPCODES


def assemble(mnemonics_str: str) -> bytes:
    mnemonics: list[str] = mnemonics_str.split(" ")
    i = 0
    bytecode = ""
    mnemonic_to_opcode = {v[0]: opcode for opcode, v in OPCODES.items()}
    while i < len(mnemonics):
        mnemonic = mnemonics[i]
        mnemonic_upper = mnemonic.upper()
        bytecode += hex(mnemonic_to_opcode[mnemonic_upper])[2:].zfill(2)

        if mnemonic_upper.startswith("PUSH"):
            mnemonic_num = int(mnemonic[4:])
            if mnemonic_num > 0:
                i += 1
                arg = mnemonics[i]
                if arg.startswith("0x"):
                    arg = arg[2:]
                else:
                    arg = hex(int(arg))[2:].zfill(mnemonic_num * 2)
                assert len(arg) == mnemonic_num * 2, "Invalid PUSH argument"
                bytecode += arg

        i += 1

    return bytes.fromhex(bytecode)



import json
import re
import sys
from pathlib import Path

"""
Usage:
$ erever trace --rpc-url $RPC_MAINNET --tx <tx hash> --output-json > ~/tmp/erever.json
$ cast run <tx hash> --trace-printer -q > ~/tmp/cast.txt
$ python test/diff_cast.py ~/tmp/erever.json ~/tmp/cast.txt
"""

if len(sys.argv) != 3:
    print("Usage: python diff_cast.py <erever output> <cast output>")
    exit()

erever_output_file = Path(sys.argv[1])
cast_output_file = Path(sys.argv[2])

erever_output = json.load(erever_output_file.open())
trace_logs = erever_output["trace_logs"]
# {'mnemonic_raw': 'SWAP2', 'mnemonic': 'SWAP', 'input': [408, 308, 308], 'stack_after_execution': [3055364086, 249, 1097077688018008265106216665536940668749033598146, 10000000000000000000000, 132, 96, 308, 308, 408]}

# cast_output_file is a big file, so we don't want to read it all into memory
with cast_output_file.open() as f:
    # Example: depth:1, PC:0, gas:0x1ba2740(28976960), OPCODE: "PUSH1"(96)  refund:0x0(0) Stack:[], Data size:0, Data: 0x
    pattern = re.compile(
        r"depth:(\d+), PC:(\d+), gas:(0x[0-9a-f]+)\((\d+)\), OPCODE: \"([A-Z0-9]+)\"\((\d+)\)  refund:(0x[0-9a-f]+)\((\d+)\) Stack:\[(.*)\], Data size:(\d+), Data: 0x([0-9a-f]*)"
    )
    trace_log_i = 0
    for line in f:
        line = line.strip()
        if line.startswith("SM CALL:"):
            print(line)
            continue
        result = pattern.match(line)
        assert result is not None
        depth = int(result.group(1))
        pc = int(result.group(2))
        gas = int(result.group(4))
        mnemonic_raw = result.group(5).replace("SHA3", "KECCAK256")
        refund = int(result.group(8))
        stack = [int(x.replace("_U256", ""), 16) for x in (result.group(9).split(",") if result.group(9) != "" else [])]
        data_size = int(result.group(10))
        data = result.group(11)
        trace_log = trace_logs[trace_log_i]

        erever_stack = trace_log["stack_before_execution"]
        erever_memory = trace_log["memory_before_execution"]
        erever_gas = trace_log["gas"]
        erever_depth = trace_log["depth"]

        print(
            "erever",
            f"depth:{erever_depth}",
            len(erever_memory) // 2,
            erever_gas,
            trace_log["mnemonic_raw"],
            erever_stack[::-1],
        )
        print("cast  ", f"depth:{depth}", len(data) // 2, gas, mnemonic_raw, stack[::-1])
        # print(line)
        print()

        if mnemonic_raw != trace_log["mnemonic_raw"]:
            print("mnemonic mismatch")
            break
        # if stack != erever_stack:
        #     print("stack mismatch")
        #     break
        # if gas != erever_gas:
        #     print("gas mismatch")
        #     break
        if data != erever_memory:
            print("erever mem", erever_memory)
            print("cast mem  ", data)
            print("memory mismatch")
            break
        if depth != erever_depth:
            print("depth mismatch")
            break

        trace_log_i += 1



from erever.context import Context
from erever.disassemble import disassemble


def test_add() -> None:
    context = Context(
        bytecode=bytes.fromhex("600a600a017fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff600101")
    )
    result = disassemble(context, trace=True, return_trace_logs=True, silent=True)
    assert result.stack_after_execution.stack == [20, 0]



import string

import rlp
from eth_typing import ChecksumAddress
from web3 import Web3

from .colors import Colors
from .types import AddressInt

UINT256_MAX = (1 << 256) - 1
SIGN_MASK = 1 << 255

PRINTABLE = string.printable[:-5]
TAB_SIZE = 4


def uint256(x: int) -> int:
    return x & UINT256_MAX


def int256(x: int) -> int:
    x &= UINT256_MAX
    if x & SIGN_MASK:
        return -((1 << 256) - x)
    else:
        return x


def pad(hex_number: str, n: int) -> str:
    if hex_number[:2] == "0x":
        hex_number = hex_number[2:]
    hex_number = "0x" + "0" * (n - len(hex_number)) + hex_number
    return hex_number


def pad_even(hex_number: str) -> str:
    if hex_number[:2] == "0x":
        hex_number = hex_number[2:]
    n = len(hex_number) + len(hex_number) % 2
    return pad(hex_number, n)


def decode_printable_with_color(
    hex_string: str,
    l_for_colorize: int | None = None,
    r_for_colorize: int | None = None,
) -> str:
    if hex_string[:2] == "0x":
        hex_string = hex_string[2:]

    decoded = ""
    for i in range(0, len(hex_string), 2):
        c = chr(int(hex_string[i : i + 2], 16))
        printable = c in PRINTABLE
        if not printable:
            c = "."
        if l_for_colorize is not None and r_for_colorize is not None and l_for_colorize <= i // 2 < r_for_colorize:
            c = Colors.GREEN + c + Colors.ENDC
        elif not printable:
            c = Colors.GRAY + "." + Colors.ENDC
        decoded += c
    return decoded


def is_invocation_mnemonic(mnemonic: str) -> bool:
    return mnemonic in [
        "CALL",
        "CALLCODE",
        "DELEGATECALL",
        "STATICCALL",
        "CREATE",
        "CREATE2",
        "SELFDESTRUCT",
        "LOG",
    ]


def int_to_check_sum_address(x: int) -> ChecksumAddress:
    return Web3.to_checksum_address(pad(hex(x), 40))


def is_overlapping(l1: int, r1: int, l2: int, r2: int) -> bool:
    return max(l1, l2) < min(r1, r2)


def compute_contract_address(address: AddressInt, nonce: int) -> AddressInt:
    return int.from_bytes(Web3.keccak(rlp.encode([address.to_bytes(20, "big"), nonce]))[-20:], "big")



from Crypto.Util.number import bytes_to_long

from .colors import Colors
from .types import Gas
from .utils import decode_printable_with_color, is_overlapping


class Memory:
    memory: list[int]
    mstore_l_for_colorize: int | None
    mstore_r_for_colorize: int | None
    MAX_MEMORY_INDEX = 1 << 24

    def __init__(self) -> None:
        self.memory = []

        self.mstore_l_for_colorize = None
        self.mstore_r_for_colorize = None

    def extend(self, size: int) -> Gas:
        if size > Memory.MAX_MEMORY_INDEX:
            raise Exception(f"Memory size too large: {size}")
        if size % 0x20 > 0:
            size += 0x20 - size % 0x20
        if len(self.memory) >= size:
            return 0
        gas = self.calculate_gas_extend_memory(size)
        self.memory += [0] * (size - len(self.memory))
        return gas

    def get_as_bytes(self, offset: int, size: int) -> bytes:
        self.extend(offset + size)
        return bytes(self.memory[offset : offset + size])

    def get_as_hex(self, offset: int, size: int) -> str:
        return self.get_as_bytes(offset, size).hex()

    def store8(self, offset: int, value: int) -> None:
        assert value < 0x100
        self.extend(offset + 1)
        self.memory[offset] = value

        self.mstore_l_for_colorize = offset
        self.mstore_r_for_colorize = offset + 1

    def store256(self, offset: int, value: int) -> Gas:
        value_bytes = value.to_bytes(32, "big")
        r = offset + 32
        allocation_gas = self.extend(r)
        for i, b in enumerate(value_bytes):
            self.memory[offset + i] = b

        self.mstore_l_for_colorize = offset
        self.mstore_r_for_colorize = r
        return allocation_gas

    def store(self, offset: int, value: bytes) -> Gas:
        r = offset + len(value)
        allocation_gas = self.extend(r)
        for i, b in enumerate(value):
            self.memory[offset + i] = b
        for i, b in enumerate(value):
            self.memory[offset + i] = b

        self.mstore_l_for_colorize = offset
        self.mstore_r_for_colorize = r
        return allocation_gas

    def load(self, offset: int) -> int:
        self.extend(offset + 32)
        return bytes_to_long(bytes(self.memory[offset : offset + 32]))

    def to_string(self, line_length: int = 0x20, memory_range: list[tuple[int, int]] | None = None) -> list[str]:
        memory_length = len(self.memory)
        memory_hex = bytes(self.memory).hex()
        ret = []
        ret_lefts = []

        def zero_to_gray(s: str) -> str:
            ret = ""
            for i in range(0, len(s), 2):
                b = s[i : i + 2]
                if b == "00":
                    ret += Colors.GRAY + b + Colors.ENDC
                else:
                    ret += b
            return ret

        if memory_range is None:
            memory_range = [(0, memory_length)]
        addrs = []
        hr_indices = []
        adding = False
        for i in range(0, memory_length, line_length):
            addr_l = i
            addr_r = i + line_length
            if any(is_overlapping(addr_l, addr_r, left, right) for left, right in memory_range):
                ret.append(memory_hex[2 * i : 2 * (i + line_length)])
                ret_lefts.append(i)
                addrs.append(hex(addr_l))
                adding = True
            else:
                if adding:
                    hr_indices.append(len(ret))
                adding = False

        decoded_lines = []

        if (
            self.mstore_l_for_colorize is not None
            and self.mstore_r_for_colorize is not None
            and self.mstore_l_for_colorize != self.mstore_r_for_colorize
        ):
            _l_i = self.mstore_l_for_colorize // line_length
            l_j = 2 * (self.mstore_l_for_colorize % line_length)
            r_i = self.mstore_r_for_colorize // line_length
            r_j = 2 * (self.mstore_r_for_colorize % line_length)

            if r_j == 0:
                r_i -= 1
                r_j = 2 * line_length

            for ret_i in range(0, len(ret)):
                line_left = ret_lefts[ret_i]
                line_right = line_left + line_length

                if not is_overlapping(
                    line_left,
                    line_right,
                    self.mstore_l_for_colorize,
                    self.mstore_r_for_colorize,
                ):
                    decoded_lines.append(decode_printable_with_color(ret[ret_i]))
                    ret[ret_i] = zero_to_gray(ret[ret_i])
                    continue

                if self.mstore_l_for_colorize <= line_left and line_right <= self.mstore_r_for_colorize:
                    decoded_lines.append(decode_printable_with_color(ret[ret_i], 0, line_length))
                    ret[ret_i] = Colors.GREEN + ret[ret_i] + Colors.ENDC
                elif line_left < self.mstore_l_for_colorize and self.mstore_r_for_colorize < line_right:
                    decoded_lines.append(decode_printable_with_color(ret[ret_i], l_j // 2, r_j // 2))
                    ret[ret_i] = (
                        zero_to_gray(ret[ret_i][:l_j])
                        + Colors.GREEN
                        + ret[ret_i][l_j:r_j]
                        + Colors.ENDC
                        + zero_to_gray(ret[ret_i][r_j:])
                    )

                elif line_left < self.mstore_l_for_colorize and line_right <= self.mstore_r_for_colorize:
                    decoded_lines.append(decode_printable_with_color(ret[ret_i], l_j // 2, line_length))
                    ret[ret_i] = zero_to_gray(ret[ret_i][:l_j]) + Colors.GREEN + ret[ret_i][l_j:] + Colors.ENDC
                elif self.mstore_l_for_colorize <= line_left and self.mstore_r_for_colorize < line_right:
                    decoded_lines.append(decode_printable_with_color(ret[ret_i], 0, r_j // 2))
                    ret[ret_i] = Colors.GREEN + ret[ret_i][:r_j] + Colors.ENDC + zero_to_gray(ret[ret_i][r_j:])
                else:
                    assert (
                        False
                    ), f"Unreachable {line_left} {line_right} {self.mstore_l_for_colorize} {self.mstore_r_for_colorize}"
            self.mstore_l_for_colorize = None
            self.mstore_r_for_colorize = None
        else:
            for i in range(0, len(ret)):
                decoded_lines.append(decode_printable_with_color(ret[i]))
                ret[i] = zero_to_gray(ret[i])

        # TODO: cleanup
        new_ret = []
        show_decode = True
        assert len(ret) == len(decoded_lines), f"{len(ret)} != {len(decoded_lines)}"
        for i in range(0, len(ret)):
            if i in hr_indices:
                new_ret.append("-" * 10)
            if show_decode:
                new_ret.append(ret[i] + " | " + decoded_lines[i] + " | " + addrs[i])
            else:
                new_ret.append(ret[i] + " | " + addrs[i])
        ret = new_ret

        return ret

    def calculate_gas_extend_memory(self, size: int) -> Gas:
        return Memory.calculate_memory_gas_cost(size) - Memory.calculate_memory_gas_cost(len(self.memory))

    @staticmethod
    def calculate_memory_gas_cost(size: int) -> Gas:
        GAS_MEMORY = 3
        size_in_words = (size + 31) // 32
        linear_cost = size_in_words * GAS_MEMORY
        quadratic_cost = size_in_words**2 // 512
        total_gas_cost = linear_cost + quadratic_cost
        return total_gas_cost



from .utils import UINT256_MAX


class Storage:
    storage: dict[int, int]

    def __init__(self) -> None:
        self.storage = {}

    def load(self, key: int) -> int:
        return self.storage.get(key, 0)

    def store(self, key: int, value: int) -> None:
        assert 0 <= value <= UINT256_MAX
        self.storage[key] = value

    def has(self, key: int) -> bool:
        return key in self.storage



from collections import deque
from copy import deepcopy

from Crypto.Util.number import bytes_to_long

from .colors import Colors
from .context import Context
from .node import Node
from .opcodes import OPCODES
from .symbolic_stack import SymbolicStack
from .utils import TAB_SIZE, UINT256_MAX, pad

Gadget = tuple[int, Node | None, int, SymbolicStack, list[Node], list[tuple[Node, int, bool]]]


def disassemble_symbolic(
    context: Context,
    entrypoint: int = 0x00,
    show_symbolic_stack: bool = False,
    max_steps: int = UINT256_MAX,
    hide_pc: bool = False,
    hide_instructions_with_no_stack_output: bool = False,
    show_opcodes: bool = False,
    silent: bool = False,
    return_gadget_list: bool = False,
) -> list[Gadget] | None:
    class State:
        context: Context
        stack: SymbolicStack
        pc: int

        steps: int
        conditions: list[tuple[Node, int, bool]]
        data_changes: list[Node] = []
        jumped_from: int | None
        jumped: bool | None

        def __init__(self, context: Context, entrypoint: int = 0x00) -> None:
            self.context = context
            self.stack = SymbolicStack()
            self.pc = entrypoint

            self.steps = 0
            self.conditions = []  # [(condition, pc, is_met: bool)]
            self.data_changes = []
            self.jumped_from = None
            self.jumped = None

        def hash(self) -> int:
            # Contexts are not changed, so they can be ignored
            return hash((self.pc, self.stack.to_string()))

    initial_state = State(context, entrypoint)
    queue: deque[State] = deque()
    queue.append(initial_state)
    hashes = set()
    gadget_list: list[Gadget] = []

    LOCATION_PAD_N = len(hex(len(context.bytecode))[2:])

    while len(queue) > 0:
        state: State = queue.popleft()
        if state.hash() in hashes:
            continue
        hashes.add(state.hash())
        context = state.context
        stack = state.stack

        if not silent:
            print(
                f"\n{Colors.BOLD}{pad(hex(state.pc), LOCATION_PAD_N)}{Colors.ENDC}",
                end="",
            )
            if state.jumped_from is not None:
                if state.jumped:
                    print(f" ({Colors.GREEN}<- {pad(hex(state.jumped_from), LOCATION_PAD_N)}{Colors.ENDC})")
                else:
                    print(f" ({Colors.RED}<- {pad(hex(state.jumped_from), LOCATION_PAD_N)}{Colors.ENDC})")
                for condition, pc, is_met in state.conditions:
                    if is_met:
                        print(
                            f"  {Colors.GREEN} {pad(hex(pc), LOCATION_PAD_N)}{Colors.ENDC}: {condition} {Colors.GREEN}== true{Colors.ENDC}"
                        )
                    else:
                        print(
                            f"  {Colors.RED} {pad(hex(pc), LOCATION_PAD_N)}{Colors.ENDC}: {condition} {Colors.RED}== false{Colors.ENDC}"
                        )
            else:
                print()

        pc = state.pc
        while pc < len(context.bytecode):
            next_pc = pc + 1
            value = context.bytecode[pc]
            if value in OPCODES:
                (
                    mnemonic_raw,
                    stack_input_count,
                    stack_output_count,
                    _base_gas,
                    _description,
                    _stack_input_names,
                ) = OPCODES[value]
            else:
                mnemonic_raw = f"{Colors.YELLOW}0x{value:02x} (?){Colors.ENDC}"
                stack_input_count = 0
                stack_output_count = 0
                _description = None

            if mnemonic_raw.startswith("PUSH"):
                mnemonic_num = int(mnemonic_raw[4:])
                push_v = bytes_to_long(context.bytecode[pc + 1 : pc + 1 + mnemonic_num])
                next_pc = pc + 1 + mnemonic_num
                mnemonic = mnemonic_raw[:4]
            elif mnemonic_raw.startswith("DUP"):
                mnemonic_num = int(mnemonic_raw[3:])
                mnemonic = mnemonic_raw[:3]
            elif mnemonic_raw.startswith("SWAP"):
                mnemonic_num = int(mnemonic_raw[4:])
                mnemonic = mnemonic_raw[:4]
            elif mnemonic_raw.startswith("LOG"):
                mnemonic_num = int(mnemonic_raw[3:])
                mnemonic = mnemonic_raw[:3]
            else:
                mnemonic_num = 0
                mnemonic = mnemonic_raw

            input: list[Node] = [stack.pop() for _ in range(stack_input_count)]
            end = False
            match mnemonic:
                # スタックの操作はここに。操作しないものはNodeの__repr__に。
                case "STOP":
                    end = True
                case (
                    "CALLDATACOPY"
                    | "CODECOPY"
                    | "EXTCODECOPY"
                    | "RETURNDATACOPY"
                    | "MSTORE"
                    | "MSTORE8"
                    | "SSTORE"
                    | "CALL"
                    | "CREATE"
                    | "CALLCODE"
                    | "DELEGATECALL"
                    | "CREATE2"
                ):
                    state.data_changes.append(Node(mnemonic, input, mnemonic_num, stack_input_count))
                    assert stack_output_count <= 1
                    if stack_output_count == 1:
                        stack.push(Node(mnemonic, input))
                case "PUSH":
                    stack.push(Node("uint256", push_v))
                case "DUP":
                    stack.extend(input[::-1] + [input[mnemonic_num - 1]])
                case "SWAP":
                    top = input[0]
                    input[0] = input[mnemonic_num]
                    input[mnemonic_num] = top
                    stack.extend(input[::-1])
                case "RETURN" | "REVERT" | "INVALID" | "SELFDESTRUCT":
                    end = True
                case _:
                    assert stack_output_count <= 1
                    if stack_output_count == 1:
                        stack.push(Node(mnemonic, input))

            if hide_instructions_with_no_stack_output and stack_output_count == 0:
                pass
            elif not silent:
                if not hide_pc:
                    print(f"{pad(hex(pc), LOCATION_PAD_N)}: ", end="")
                if show_opcodes:
                    print(
                        f"{Colors.GRAY}(0x{context.bytecode[pc:pc+1].hex()}){Colors.ENDC} ",
                        end="",
                    )

                if mnemonic == "PUSH":
                    print(
                        Node(
                            mnemonic,
                            [Node("uint256", push_v)],
                            mnemonic_num,
                            stack_input_count,
                        ),
                        end="",
                    )
                else:
                    res = str(Node(mnemonic, input, mnemonic_num, stack_input_count))
                    if res[0] == "(":
                        print(res[1:-1], end="")
                    else:
                        print(res, end="")

                print()

                if show_symbolic_stack:
                    print(f"{'stack'.rjust(TAB_SIZE * 2)}{' ' * TAB_SIZE}{stack.to_string()}")

            if mnemonic == "JUMP" and input[0].type != "uint256":
                gadget_list.append(
                    (
                        pc,
                        input[0],
                        state.stack.var_n,
                        state.stack,
                        state.data_changes,
                        state.conditions,
                    )
                )
                break
            if mnemonic == "JUMPI" and input[0].type != "uint256":
                state.steps += 1
                state_not_jumped = deepcopy(state)
                state_not_jumped.pc = next_pc
                state_not_jumped.jumped_from = pc
                state_not_jumped.jumped = False
                state_not_jumped.conditions.append((input[1], pc, False))
                queue.append(state_not_jumped)
                state_jumped = state
                state_jumped.conditions.append((input[1], pc, True))
                gadget_list.append(
                    (
                        pc,
                        input[0],
                        state.stack.var_n,
                        state.stack,
                        state.data_changes,
                        state_jumped.conditions,
                    )
                )
                break
            if mnemonic == "JUMP" and input[0].type == "uint256":
                assert isinstance(input[0].value, int)
                next_pc = input[0].value
            if mnemonic == "JUMPI" and input[0].type == "uint256":
                assert isinstance(input[0].value, int)
                state.steps += 1
                if input[1].evaluate() != "false":
                    state_jumped = deepcopy(state)
                    state_jumped.pc = input[0].value
                    state_jumped.jumped_from = pc
                    state_jumped.jumped = True
                    state_jumped.conditions.append((input[1], pc, True))
                    queue.append(state_jumped)
                if input[1].evaluate() != "true":
                    state_not_jumped = state
                    state_not_jumped.pc = next_pc
                    state_not_jumped.jumped_from = pc
                    state_not_jumped.jumped = False
                    state_not_jumped.conditions.append((input[1], pc, False))
                    queue.append(state_not_jumped)
                break
            if mnemonic == "STOP" or mnemonic == "RETURN":
                gadget_list.append(
                    (
                        pc,
                        None,
                        state.stack.var_n,
                        state.stack,
                        state.data_changes,
                        state.conditions,
                    )
                )
                break

            state.steps += 1
            if state.steps >= max_steps:
                if not silent:
                    print("The maximum number of steps has been reached.")
                break

            if end:
                break

            pc = next_pc

    if return_gadget_list:
        return gadget_list
    else:
        return None



import argparse

from Crypto.Util.number import bytes_to_long
from web3 import HTTPProvider, Web3
from web3.types import BlockData, TxData

from .precompiled_contracts import PRECOMPILED_CONTRACTS
from .storage import Storage
from .types import AddressInt, Gas
from .utils import UINT256_MAX, int_to_check_sum_address

StateDict = dict[str, dict[str, str | int | dict[str, str]]]


class State:
    w3: Web3 | None
    balances: dict[AddressInt, int]
    codes: dict[AddressInt, bytes]
    storages: dict[AddressInt, Storage]
    original_storages: dict[AddressInt, Storage]
    address_access_set: set[AddressInt]
    block_number: int

    def __init__(
        self,
        block_number: int,
        rpc_url: str | None = None,
        state_dict: StateDict | None = None,
    ) -> None:
        self.w3 = Web3(HTTPProvider(rpc_url)) if rpc_url else None
        self.block_number = block_number
        self.storages = {}
        self.original_storages = {}
        self.codes = {address: b"" for address in PRECOMPILED_CONTRACTS}
        self.address_access_set = set([address for address in PRECOMPILED_CONTRACTS])
        self.balances = {}

        if state_dict:
            if "balance" in state_dict:
                for address, balance in state_dict["balance"].items():
                    address_int = int(address, 16)
                    if isinstance(balance, str):
                        balance_int = int(balance, 16 if balance.startswith("0x") else 10)
                    elif isinstance(balance, int):
                        balance_int = balance
                    else:
                        raise Exception("Invalid balance")
                    assert 0 <= balance_int < UINT256_MAX
                    self.balances[address_int] = balance_int
            if "code" in state_dict:
                for address, code in state_dict["code"].items():
                    assert isinstance(code, str)
                    self.codes[int(address, 16)] = bytes.fromhex(code.replace("0x", "").replace(" ", ""))
            if "storage" in state_dict:
                for address, d in state_dict["storage"].items():
                    address_int = int(address, 16)
                    self.storages[address_int] = Storage()
                    assert isinstance(d, dict)
                    for slot, value in d.items():
                        assert isinstance(value, str)
                        slot_int = int(slot, 16 if slot.startswith("0x") else 10)
                        value_int = int(value, 16 if value.startswith("0x") else 10)
                        self.storages[address_int].store(slot_int, value_int)

    def get_balance(self, address: AddressInt) -> int:
        if address in self.balances:
            return self.balances[address]
        elif self.w3:
            self.balances[address] = self.w3.eth.get_balance(int_to_check_sum_address(address), self.block_number)
            return self.balances[address]
        else:
            return 0

    def set_balance(self, address: AddressInt, balance: int) -> None:
        self.balances[address] = balance

    def get_code(self, address: AddressInt) -> bytes:
        if address in self.codes:
            code = self.codes[address]
            return code

        if self.w3:
            code = bytes(self.w3.eth.get_code(int_to_check_sum_address(address), self.block_number))
            self.codes[address] = code
            return code

        return b""

    def set_code(self, address: AddressInt, code: bytes) -> None:
        self.codes[address] = code

    def get_storage_at(self, address: AddressInt, slot: int) -> tuple[int, Gas]:
        GAS_WARM_COLD_DIFF = 2100 - 100

        if address not in self.storages:
            self.storages[address] = Storage()
        if address not in self.original_storages:
            self.original_storages[address] = Storage()

        storage = self.storages[address]
        original_storage = self.original_storages[address]

        if not original_storage.has(slot):
            if self.w3:
                original_storage.store(
                    slot,
                    bytes_to_long(
                        self.w3.eth.get_storage_at(int_to_check_sum_address(address), slot, self.block_number)
                    ),
                )
            else:
                original_storage.store(slot, 0)

        original_value = original_storage.load(slot)
        warm = storage.has(slot)
        if not warm:
            storage.store(slot, original_value)
        current_value = storage.load(slot)

        return (current_value, 0 if warm else GAS_WARM_COLD_DIFF)

    def set_storage_at(self, address: AddressInt, slot: int, value: int) -> tuple[Gas, Gas]:
        if address not in self.storages:
            self.storages[address] = Storage()
        if address not in self.original_storages:
            self.original_storages[address] = Storage()

        storage = self.storages[address]
        original_storage = self.original_storages[address]

        if not original_storage.has(slot):
            if self.w3:
                original_storage.store(
                    slot,
                    bytes_to_long(
                        self.w3.eth.get_storage_at(int_to_check_sum_address(address), slot, self.block_number)
                    ),
                )
            else:
                original_storage.store(slot, 0)

        original_value = original_storage.load(slot)
        warm = storage.has(slot)
        if not warm:
            storage.store(slot, original_value)
        current_value = storage.load(slot)

        # https://www.evm.codes/#55?fork=shanghai
        if value == current_value:
            base_dynamic_gas = 100
        elif current_value == original_value:
            if original_value == 0:
                base_dynamic_gas = 20000
            else:
                base_dynamic_gas = 2900
        else:
            base_dynamic_gas = 100

        gas_refunds = 0
        if value != current_value:
            if current_value == original_value:
                if original_value != 0 and value == 0:
                    gas_refunds += 4800
            else:
                if original_value != 0:
                    if current_value == 0:
                        gas_refunds -= 4800
                    elif value == 0:
                        gas_refunds += 4800
                if value == original_value:
                    if original_value == 0:
                        gas_refunds += 19900
                    else:
                        if warm:
                            gas_refunds += 5000 - 2100 - 100
                        else:
                            gas_refunds += 4900
        storage.store(slot, value)

        # print(address, slot, " ", base_dynamic_gas, gas_refunds, value, current_value, original_value, warm, file=sys.stderr)

        gas = base_dynamic_gas - 100
        # gas refunds are applied at the end of the transaction
        return (gas, gas_refunds)


class Context:
    DEFAULT_BYTECODE = b""
    DEFAULT_ADDRESS = 0xADD2E55
    DEFAULT_ORIGIN = 0
    DEFAULT_CALLER = 0
    DEFAULT_CALLVALUE = 0
    DEFAULT_CALLDATA = b""
    DEFAULT_GASPRICE = 0
    DEFAULT_COINBASE = 0
    DEFAULT_TIMESTAMP = 0
    DEFAULT_NUMBER = 0
    DEFAULT_DIFFICULTY = 0
    DEFAULT_GASLIMIT = 0
    DEFAULT_CHAINID = 1
    DEFAULT_SELFBALANCE = 0
    DEFAULT_BASEFEE = 0
    DEFAULT_GAS = UINT256_MAX

    state: State
    bytecode: bytes
    address: int
    origin: int
    caller: int
    callvalue: int
    calldata: bytes
    gasprice: int
    coinbase: int
    timestamp: int
    number: int
    difficulty: int
    gaslimit: int
    chainid: int
    selfbalance: int
    basefee: int
    gas: int

    static: bool
    return_data: bytes
    depth: int
    steps: int

    def __init__(
        self,
        bytecode: bytes = DEFAULT_BYTECODE,
        address: int = DEFAULT_ADDRESS,
        origin: int = DEFAULT_ORIGIN,
        caller: int = DEFAULT_CALLER,
        callvalue: int = DEFAULT_CALLVALUE,
        calldata: bytes = DEFAULT_CALLDATA,
        gasprice: int = DEFAULT_GASPRICE,
        coinbase: int = DEFAULT_COINBASE,
        timestamp: int = DEFAULT_TIMESTAMP,
        number: int = DEFAULT_NUMBER,
        difficulty: int = DEFAULT_DIFFICULTY,
        gaslimit: int = DEFAULT_GASLIMIT,
        chainid: int = DEFAULT_CHAINID,
        selfbalance: int = DEFAULT_SELFBALANCE,
        basefee: int = DEFAULT_BASEFEE,
        gas: int = DEFAULT_GAS,
        rpc_url: str | None = None,
        state_dict: StateDict | None = None,
    ) -> None:
        self.bytecode = bytecode
        self.address = address
        self.origin = origin
        self.caller = caller
        self.callvalue = callvalue
        self.calldata = calldata
        self.gasprice = gasprice
        self.coinbase = coinbase
        self.timestamp = timestamp
        self.number = number
        self.difficulty = difficulty
        self.gaslimit = gaslimit
        self.chainid = chainid
        self.selfbalance = selfbalance
        self.basefee = basefee
        self.gas = gas
        self.state = State(self.number, rpc_url, state_dict)

        self.static = False
        self.return_data = b""
        self.depth = 1
        self.steps = 0

    @classmethod
    def from_arg_params_with_bytecode(cls, args: argparse.Namespace, bytecode: str) -> "Context":
        return cls(
            bytecode=Context.__hex_to_bytes(bytecode),
            address=args.address,
            origin=args.origin,
            caller=args.caller,
            callvalue=args.callvalue,
            calldata=Context.__hex_to_bytes(args.calldata),
            gasprice=args.gasprice,
            coinbase=args.coinbase,
            timestamp=args.timestamp,
            number=args.number,
            difficulty=args.difficulty,
            gaslimit=args.gaslimit,
            chainid=args.chainid,
            selfbalance=args.selfbalance,
            basefee=args.basefee,
            gas=args.gas,
            rpc_url=args.rpc_url,
        )

    @classmethod
    def from_tx_hash(cls, args: argparse.Namespace) -> "Context":
        assert args.rpc_url, "RPC URL must be specified"

        w3 = Web3(HTTPProvider(args.rpc_url))
        tx: TxData = w3.eth.get_transaction(args.tx)
        tx_receipt = w3.eth.get_transaction_receipt(args.tx)

        self = cls()
        previous_block_number = tx["blockNumber"] - 1  # not includes tx
        current_block_number = tx["blockNumber"]  # includes tx
        self.state = State(previous_block_number, args.rpc_url)

        current_block: BlockData = w3.eth.get_block(current_block_number)

        # Contract Creation
        if "to" not in tx or tx["to"] is None:
            self.bytecode = bytes(tx["input"])
            self.calldata = b""
            assert tx_receipt["contractAddress"] is not None
            to_address = int(str(tx_receipt["contractAddress"]), 16)
        else:
            to_address = int(tx["to"], 16)
            self.state.address_access_set.add(to_address)
            code = self.state.get_code(to_address)
            # Contract
            if len(code) > 0:
                self.bytecode = bytes(code)
                self.calldata = bytes(tx["input"])
            # EOA
            else:
                self.bytecode = bytes(tx["input"])
                self.calldata = b""

        self.address = to_address
        self.origin = int(tx["from"], 16)
        self.caller = int(tx["from"], 16)
        self.callvalue = tx["value"]
        self.gasprice = tx["gasPrice"]
        self.coinbase = args.coinbase
        self.timestamp = current_block["timestamp"]
        self.number = current_block_number
        self.difficulty = args.difficulty
        self.gaslimit = tx["gas"]
        self.chainid = tx.get("chainId", args.chainid)
        self.selfbalance = args.selfbalance
        self.basefee = current_block.get("baseFeePerGas", args.basefee)
        self.gas = tx["gas"]
        # self.blockchash

        self.static = False
        self.return_data = b""
        self.depth = 1
        self.steps = 0
        return self

    @classmethod
    def from_contract_address(cls, args: argparse.Namespace) -> "Context":
        self = cls()
        assert args.rpc_url, "RPC URL must be specified"

        w3 = Web3(HTTPProvider(args.rpc_url))
        if args.number == 0:
            args.number = w3.eth.block_number
        if args.timestamp == 0:
            args.timestamp = w3.eth.get_block(args.number)["timestamp"]

        code = w3.eth.get_code(args.contract_address, args.number)

        self.bytecode = bytes(code)
        assert args.address in [
            None,
            Context.DEFAULT_ADDRESS,
        ], "address must not be specified"
        self.address = Context.__hex_to_int(args.contract_address)

        self.origin = args.origin
        self.caller = args.caller
        self.callvalue = args.callvalue
        self.calldata = Context.__hex_to_bytes(args.calldata)
        self.gasprice = args.gasprice
        self.coinbase = args.coinbase
        self.timestamp = args.timestamp
        self.number = args.number
        self.difficulty = args.difficulty
        self.gaslimit = args.gas
        self.chainid = args.chainid
        self.selfbalance = args.selfbalance
        self.basefee = args.basefee
        self.gas = args.gas
        # self.blockchash

        self.state = State(self.number, args.rpc_url)
        self.static = False
        self.return_data = b""
        self.depth = 1
        self.steps = 0
        return self

    @staticmethod
    def __hex_to_bytes(x: str | int) -> bytes:
        if isinstance(x, int):
            h = hex(x)
        elif isinstance(x, str):
            h = x
        h = h.replace(" ", "").replace("\n", "")
        if h.startswith("0x"):
            h = h[2:]
        return bytes.fromhex(h)

    @staticmethod
    def __hex_to_int(h: str) -> int:
        if h.startswith("0x"):
            return int(h, 16)
        else:
            return int(h)



import base58
import cbor2
import requests


def inspect_cbor(bytecode: bytes) -> None:
    if len(bytecode) < 2:
        return
    length = int.from_bytes(bytecode[-2:], "big")
    if length >= len(bytecode):
        print("CBOR: Not found")
        return
    print(f"CBOR: length = {length}")

    cbor_data = bytecode[-length - 2 : -2]
    try:
        cbor_obj = cbor2.loads(cbor_data)
    except Exception:
        print("CBOR: Failed to parse")
        return
    assert isinstance(cbor_obj, dict)
    for k, v in cbor_obj.items():
        assert isinstance(k, str)
        assert isinstance(v, bytes)
        if k == "ipfs":
            ipfs_hash = base58.b58encode(v)
            ipfs_url = f"https://ipfs.io/ipfs/{ipfs_hash.decode()}"
            print(f"CBOR: {k} = {ipfs_url}")
            response = requests.get(ipfs_url)
            if response.status_code == 200:
                response_data = response.text.encode()
                if len(response_data) > 0x100:
                    response_data = response_data[:0x100] + b"..."
                print("CBOR: ipfs content =", response_data)
        elif k == "solc":
            assert len(v) == 3
            x, y, z = v
            print(f"CBOR: {k} = {x}.{y}.{z}")
        else:
            print(f"CBOR: {k} = {v.hex()}")



from .colors import Colors
from .context import Context
from .disassemble import disassemble
from .disassemble_symbolic import disassemble_symbolic
from .utils import pad


def find_gadgets(
    context: Context,
    max_steps: int = 1000,
) -> None:
    disassembled_code = disassemble(context, silent=True).disassemble_code
    LOCATION_PAD_N = len(hex(len(context.bytecode))[2:])

    for pc, mnemonic, _push_v in disassembled_code:
        if mnemonic != "JUMPDEST":
            continue

        print(f"Checking {pad(hex(pc), LOCATION_PAD_N)}")
        gadget_list = disassemble_symbolic(
            context,
            entrypoint=pc,
            show_symbolic_stack=True,
            max_steps=max_steps,
            silent=True,
            return_gadget_list=True,
        )
        assert gadget_list is not None
        for gadget in gadget_list:
            jump_pc, jump_dst, var_n, stack, data_changes, conditions = gadget
            print(" ", "Gadget:")
            print(" ", " ", "    entry pc:", pad(hex(pc), LOCATION_PAD_N))
            print(" ", " ", "     jump pc:", pad(hex(jump_pc), LOCATION_PAD_N))
            print(" ", " ", "    jump dst:", jump_dst)
            print(" ", " ", " stack input:", var_n)
            print(" ", " ", "stack output:", len(stack), stack)
            if len(data_changes) > 0:
                print(" ", " ", "data changes:")
                for data_change in data_changes:
                    print(" " * 10, data_change)
            if len(conditions) > 0:
                print(" ", " ", "  conditions:")
                for condition, condition_pc, is_met in conditions:
                    if is_met:
                        print(
                            " " * 10
                            + f"{Colors.GREEN} {pad(hex(condition_pc), LOCATION_PAD_N)}{Colors.ENDC}: {condition} {Colors.GREEN}== true{Colors.ENDC}"
                        )
                    else:
                        print(
                            " " * 10
                            + f"{Colors.RED} {pad(hex(condition_pc), LOCATION_PAD_N)}{Colors.ENDC}: {condition} {Colors.RED}== false{Colors.ENDC}"
                        )
            print()



from dataclasses import dataclass

from .types import Gas

GAS_CALL_STIPEND = 2300
GAS_CODE_WARM_COLD_DIFF = 2600 - 100


@dataclass
class MessageCallGas:
    cost: int
    stipend: int


def calculate_message_call_gas(
    value: int,
    gas: Gas,
    gas_left: Gas,
    memory_cost: Gas,
    extra_gas: Gas,
    call_stipend: Gas = GAS_CALL_STIPEND,
) -> MessageCallGas:
    call_stipend = 0 if value == 0 else call_stipend
    if gas_left < extra_gas + memory_cost:
        return MessageCallGas(gas + extra_gas, gas + call_stipend)

    gas = min(gas, max_message_call_gas(gas_left - memory_cost - extra_gas))

    return MessageCallGas(gas + extra_gas, gas + call_stipend)


def max_message_call_gas(gas: Gas) -> Gas:
    return gas - (gas // 64)



import copy
import sys
from dataclasses import dataclass
from enum import Enum

from Crypto.Hash import SHA256, keccak
from Crypto.Util.number import bytes_to_long

from .colors import Colors
from .context import Context
from .gas import GAS_CODE_WARM_COLD_DIFF, calculate_message_call_gas
from .memory import Memory
from .opcodes import OPCODES
from .precompiled_contracts import PRECOMPILED_CONTRACTS
from .stack import Stack
from .utils import (
    SIGN_MASK,
    TAB_SIZE,
    UINT256_MAX,
    compute_contract_address,
    int256,
    is_invocation_mnemonic,
    pad,
    pad_even,
    uint256,
)


@dataclass
class TraceLog:
    mnemonic_raw: str
    mnemonic: str
    input: list[int] | int
    stack_before_execution: Stack
    memory_before_execution: Memory
    gas: int
    depth: int

    def to_dict(self) -> dict[str, int | str | list[int]]:
        log_dict: dict[str, int | str | list[int]] = {}
        log_dict["mnemonic_raw"] = self.mnemonic_raw
        log_dict["mnemonic"] = self.mnemonic
        log_dict["input"] = self.input
        log_dict["stack_before_execution"] = self.stack_before_execution.stack
        log_dict["memory_before_execution"] = bytes(self.memory_before_execution.memory).hex()
        log_dict["gas"] = self.gas
        log_dict["depth"] = self.depth
        return log_dict


DisassembleCode = list[tuple[int, str, int | None]]
DisassembleResultDict = dict[
    str,
    int | None | DisassembleCode | list[dict[str, str | int | list[int]]] | list[int] | str,
]  # TODO


@dataclass
class DisassembleResult:
    last_jump_to_address: int | None
    disassemble_code: DisassembleCode
    trace_logs: list[TraceLog]
    success: bool
    return_data: bytes
    stack_after_execution: Stack
    memory_after_execution: Memory

    def to_dict(
        self,
    ) -> DisassembleResultDict:
        result_dict: DisassembleResultDict = {}
        result_dict["last_jump_to_address"] = self.last_jump_to_address
        result_dict["disassemble_code"] = self.disassemble_code
        result_dict["trace_logs"] = [log.to_dict() for log in self.trace_logs]
        result_dict["stack_after_execution"] = self.stack_after_execution.stack
        result_dict["memory_after_execution"] = bytes(self.memory_after_execution.memory).hex()
        return result_dict


class MemoryDisplay(Enum):
    ALWAYS = "always"
    ONCHANGE = "onchange"
    OFF = "off"

    def __str__(self) -> str:
        return self.value


class StaticCallError(Exception):
    pass


def disassemble(
    context: Context,
    trace: bool = False,
    entrypoint: int = 0x00,
    max_steps: int = UINT256_MAX,
    decode_stack: bool = False,
    ignore_stack_underflow: bool = False,
    silent: bool = False,
    hide_pc: bool = False,
    show_opcodes: bool = False,
    memory_display: MemoryDisplay = MemoryDisplay.OFF,
    memory_range: list[tuple[int, int]] | None = None,
    invocation_only: bool = False,
    return_trace_logs: bool = False,
) -> DisassembleResult:
    disassembled_code: DisassembleCode = []  # (pc, mnemonic, push_v)
    stack = Stack(ignore_stack_underflow=ignore_stack_underflow)
    memory = Memory()
    success = True
    return_data = b""

    LOCATION_PAD_N = len(hex(len(context.bytecode))[2:])

    warning_messages: str = ""

    pc = entrypoint
    last_jump_to_address = None
    trace_logs = []

    if context.chainid == 1 and context.address in PRECOMPILED_CONTRACTS:
        precompiled_contract_name, precompiled_contract_gas = PRECOMPILED_CONTRACTS[context.address]
        context.gas -= precompiled_contract_gas
        data_word_size = (len(context.calldata) + 31) // 32
        match precompiled_contract_name:
            case "ecRecover":
                assert False, "ecRecover is not supported"
            case "SHA2-256":
                context.gas -= 12 * data_word_size
                return_data = SHA256.new(data=context.calldata).digest()
            case "RIPEMD-160":
                context.gas -= 120 * data_word_size
                assert False, "RIPEMD-160 is not supported"
            case "identity":
                context.gas -= 3 * data_word_size
                return_data = context.calldata
            case "modexp":
                assert False, "modexp is not supported"
            case "ecAdd":
                assert False, "ecAdd is not supported"
            case "ecMul":
                assert False, "ecMul is not supported"
            case "ecPairing":
                assert False, "ecPairing is not supported"
            case "blake2f":
                assert False, "blake2f is not supported"

    while pc < len(context.bytecode):
        instruction_message = ""
        next_pc = pc + 1
        value = context.bytecode[pc]
        if value in OPCODES:
            (
                mnemonic_raw,
                stack_input_count,
                _stack_output_count,
                base_gas,
                _description,
                stack_input_names,
            ) = OPCODES[value]
        else:
            mnemonic_raw = Colors.YELLOW + f"0x{value:02x} (?)" + Colors.ENDC
            stack_input_count = 0
            _stack_output_count = 0
            base_gas = 0
            _description = None
            stack_input_names = ()

            warning_messages += f"The mnemonic for 0x{value:02x} in {pad(hex(pc), LOCATION_PAD_N)} is not found.\n"

        if not silent:
            if not hide_pc:
                instruction_message += f"{pad(hex(pc), LOCATION_PAD_N)}: "
            if show_opcodes:
                instruction_message += f"{Colors.GRAY}(0x{context.bytecode[pc:pc+1].hex()}){Colors.ENDC} "
            if mnemonic_raw in ["JUMP", "JUMPI"]:
                instruction_message += f"{Colors.CYAN}{Colors.BOLD}{mnemonic_raw}{Colors.ENDC}"
            elif mnemonic_raw == "JUMPDEST":
                instruction_message += f"{Colors.BLUE}{Colors.BOLD}{mnemonic_raw}{Colors.ENDC}"
            else:
                instruction_message += f"{Colors.BOLD}{mnemonic_raw}{Colors.ENDC}"

        # separate instruction names from numerical values
        mnemonic = ""
        if mnemonic_raw.startswith("PUSH"):
            mnemonic_num = int(mnemonic_raw[4:])
            push_v = bytes_to_long(context.bytecode[pc + 1 : pc + 1 + mnemonic_num])
            if not silent and mnemonic_num > 0 and not invocation_only:
                instruction_message += " 0x" + context.bytecode[pc + 1 : pc + 1 + mnemonic_num].hex()
            next_pc = pc + 1 + mnemonic_num
            mnemonic = mnemonic_raw[:4]
        elif mnemonic_raw.startswith("DUP"):
            mnemonic_num = int(mnemonic_raw[3:])
            mnemonic = mnemonic_raw[:3]
        elif mnemonic_raw.startswith("SWAP"):
            mnemonic_num = int(mnemonic_raw[4:])
            mnemonic = mnemonic_raw[:4]
        elif mnemonic_raw.startswith("LOG"):
            mnemonic_num = int(mnemonic_raw[3:])
            mnemonic = mnemonic_raw[:3]
        else:
            mnemonic_num = 0
            mnemonic = mnemonic_raw
        assert mnemonic != ""

        if mnemonic == "PUSH":
            disassembled_code.append((pc, mnemonic_raw, push_v))
        else:
            disassembled_code.append((pc, mnemonic_raw, None))

        break_flag = False

        if trace:
            if return_trace_logs:
                copied_stack = copy.deepcopy(stack)
                copied_memory = copy.deepcopy(memory)

            input: list[int] = []
            for _ in range(stack_input_count):
                input.append(stack.pop())

            if return_trace_logs:
                trace_logs.append(
                    TraceLog(
                        mnemonic_raw,
                        mnemonic,
                        copy.deepcopy(input) if mnemonic != "PUSH" else push_v,
                        copied_stack,
                        copied_memory,
                        context.gas,
                        context.depth,
                    )
                )

            context.gas -= base_gas

            if not silent:
                if len(stack_input_names) > 0:
                    instruction_message += "("
                    if mnemonic == "DUP":
                        if mnemonic_num >= 2:
                            instruction_message += "..., "
                        instruction_message += f"{pad_even(hex(input[-1]))}"
                    elif mnemonic == "SWAP":
                        if mnemonic_num >= 2:
                            instruction_message += f"{pad_even(hex(input[0]))}, ..., {pad_even(hex(input[-1]))}"
                        else:
                            instruction_message += f"{pad_even(hex(input[0]))}, {pad_even(hex(input[-1]))}"
                    else:
                        for i, name in enumerate(stack_input_names):
                            if i > 0:
                                instruction_message += ", "
                            if name != "":
                                instruction_message += f"{name}:"
                            instruction_message += f"{pad_even(hex(input[i]))}"
                    instruction_message += ")"

            try:
                if context.static and mnemonic in [
                    "SSTORE",
                    "CREATE",
                    "CREATE2",
                    "CALLCODE",  # TODO
                    "SELFDESTRUCT",
                ]:
                    raise StaticCallError
                match mnemonic:
                    case "STOP":
                        break_flag = True
                    case "ADD":
                        stack.push(uint256(input[0] + input[1]))
                    case "MUL":
                        stack.push(uint256(input[0] * input[1]))
                    case "SUB":
                        stack.push(uint256(input[0] - input[1]))
                    case "DIV":
                        if input[1] == 0:
                            stack.push(0)
                        else:
                            stack.push(input[0] // input[1])
                    case "SDIV":
                        if input[1] == 0:
                            stack.push(0)
                        else:
                            stack.push(int256(input[0]) // int256(input[1]))
                    case "MOD":
                        stack.push(input[0] % input[1])
                    case "SMOD":
                        stack.push(int256(input[0]) % int256(input[1]))
                    case "ADDMOD":
                        stack.push((input[0] + input[1]) % input[2])
                    case "MULMOD":
                        stack.push((input[0] * input[1]) % input[2])
                    case "EXP":
                        a, exp = input
                        stack.push(uint256(a**exp))
                        exp_byte_size = (exp.bit_length() + 7) // 8
                        context.gas -= 50 * exp_byte_size
                    case "SIGNEXTEND":
                        byte_num, value = input
                        if byte_num > 31:
                            stack.push(input[1])
                        else:
                            bits = (byte_num + 1) * 8
                            mask = (1 << bits) - 1
                            value &= mask
                            sign = value >> (bits - 1)
                            if sign:
                                stack.push((1 << 256) - ((1 << bits) - value))
                            else:
                                stack.push(value)
                    case "LT":
                        stack.push(int(input[0] < input[1]))
                    case "GT":
                        stack.push(int(input[0] > input[1]))
                    case "SLT":
                        stack.push(int(int256(input[0]) < int256(input[1])))
                    case "SGT":
                        stack.push(int(int256(input[0]) > int256(input[1])))
                    case "EQ":
                        stack.push(int(input[0] == input[1]))
                    case "ISZERO":
                        stack.push(int(input[0] == 0))
                    case "AND":
                        stack.push(input[0] & input[1])
                    case "OR":
                        stack.push(input[0] | input[1])
                    case "XOR":
                        stack.push(input[0] ^ input[1])
                    case "NOT":
                        stack.push(UINT256_MAX ^ input[0])
                    case "BYTE":
                        if input[0] < 32:
                            stack.push(input[1].to_bytes(32, "big")[input[0]])
                        else:
                            stack.push(0)
                    case "SHL":
                        stack.push(uint256(input[1] << input[0]))
                    case "SHR":
                        stack.push(input[1] >> input[0])
                    case "SAR":
                        if input[1] & SIGN_MASK:
                            stack.push(((1 << 256) + (int256(input[1]) >> input[0])))
                        else:
                            stack.push(input[1] >> input[0])
                    case "KECCAK256":
                        input_data = memory.get_as_bytes(input[0], input[1])
                        if not silent:
                            instruction_message += f"\n{'input'.rjust(TAB_SIZE * 2)}{' ' * TAB_SIZE}{input_data.hex()}"
                        GAS_KECCAK256_WORD = 6
                        context.gas -= GAS_KECCAK256_WORD * ((len(input_data) + 31) // 32)
                        stack.push(bytes_to_long(keccak.new(digest_bits=256, data=input_data).digest()))
                    case "ADDRESS":
                        stack.push(context.address)
                    case "BALANCE":
                        address = input[0]
                        if address not in context.state.address_access_set:
                            context.gas -= GAS_CODE_WARM_COLD_DIFF
                            context.state.address_access_set.add(address)
                        stack.push(address)
                    case "ORIGIN":
                        stack.push(context.origin)
                    case "CALLER":
                        stack.push(context.caller)
                    case "CALLVALUE":
                        stack.push(context.callvalue)
                    case "CALLDATALOAD":
                        if len(context.calldata) < input[0]:
                            stack.push(0)
                        else:
                            stack.push(bytes_to_long((context.calldata[input[0] :] + b"\x00" * 32)[:32]))
                    case "CALLDATASIZE":
                        stack.push(len(context.calldata))
                    case "CALLDATACOPY":
                        offset = input[1]
                        size = input[2]
                        if offset > len(context.calldata):
                            context.gas -= memory.store(input[0], b"\x00" * size)
                        elif offset + size > len(context.calldata):
                            context.gas -= memory.store(
                                input[0],
                                context.calldata[offset:] + b"\x00" * (offset + size - len(context.calldata)),
                            )
                        else:
                            context.gas -= memory.store(input[0], context.calldata[offset : offset + size])
                        context.gas -= 3 * ((size + 31) // 32)
                    case "CODESIZE":
                        stack.push(len(context.bytecode))
                    case "CODECOPY":
                        # TODO: bound
                        memory.store(input[0], context.bytecode[input[1] : input[1] + input[2]])
                    case "GASPRICE":
                        stack.push(context.callvalue)
                    case "EXTCODESIZE":
                        address = input[0]
                        if address not in context.state.address_access_set:
                            context.gas -= GAS_CODE_WARM_COLD_DIFF
                            context.state.address_access_set.add(address)
                        code = context.state.get_code(address)
                        stack.push(len(code))
                    case "EXTCODECOPY":
                        address, mem_offset, offset, size = input
                        if address not in context.state.address_access_set:
                            context.gas -= GAS_CODE_WARM_COLD_DIFF
                            context.state.address_access_set.add(address)
                        code = context.state.get_code(address)
                        # TODO: bounds, gas
                        context.gas -= memory.store(
                            mem_offset,
                            code[offset : offset + size] + b"\x00" * max(0, (offset + size - len(code))),
                        )
                    case "RETURNDATASIZE":
                        stack.push(len(context.return_data))
                    case "RETURNDATACOPY":
                        dest_offset, offset, size = input
                        allocation_gas = memory.store(dest_offset, context.return_data[offset : offset + size])
                        context.gas -= allocation_gas
                        context.gas -= 3 * ((size + 31) // 32)
                    case "EXTCODEHASH":
                        address = input[0]
                        if address not in context.state.address_access_set:
                            context.gas -= GAS_CODE_WARM_COLD_DIFF
                            context.state.address_access_set.add(address)
                        # TODO: empty account
                        code = context.state.get_code(address)
                        if len(code) == 0:
                            codehash = 0
                        else:
                            codehash = bytes_to_long(keccak.new(digest_bits=256, data=code).digest())
                        stack.push(codehash)
                    case "BLOCKHASH":
                        assert False, "BLOCKHASH is not supported"
                    case "COINBASE":
                        stack.push(context.coinbase)
                    case "TIMESTAMP":
                        stack.push(context.timestamp)
                    case "NUMBER":
                        stack.push(context.number)
                    case "PREVRANDAO":  # TODO
                        stack.push(context.difficulty)
                    case "GASLIMIT":
                        stack.push(context.gaslimit)
                    case "CHAINID":
                        stack.push(context.chainid)
                    case "SELFBALANCE":
                        stack.push(context.selfbalance)
                    case "BASEFEE":
                        stack.push(context.basefee)
                    case "POP":
                        pass
                    case "MLOAD":
                        stack.push(memory.load(input[0]))
                    case "MSTORE":
                        context.gas -= memory.store256(input[0], input[1])
                    case "MSTORE8":
                        memory.store8(input[0], input[1])
                    case "SLOAD":
                        value, gas = context.state.get_storage_at(context.address, input[0])
                        stack.push(value)
                        context.gas -= gas
                    case "SSTORE":
                        gas = context.state.set_storage_at(context.address, input[0], input[1])[0]
                        context.gas -= gas
                    case "JUMP":
                        assert OPCODES[context.bytecode[input[0]]][0] == "JUMPDEST", "Invalid jump destination"
                        assert isinstance(input[0], int)
                        next_pc = input[0]
                        last_jump_to_address = input[0]
                    case "JUMPI":
                        assert OPCODES[context.bytecode[input[0]]][0] == "JUMPDEST", "Invalid jump destination"
                        assert isinstance(input[0], int)
                        if input[1] != 0:
                            next_pc = input[0]
                        last_jump_to_address = input[0]
                    case "PC":
                        stack.push(pc)
                    case "MSIZE":
                        stack.push(len(memory.memory))
                    case "GAS":
                        stack.push(context.gas)
                    case "JUMPDEST":
                        pass
                    case "MCOPY":
                        # TODO: strict
                        memory.store(input[0], memory.get_as_bytes(input[1], input[2]))
                    case "PUSH":
                        stack.push(push_v)
                    case "DUP":
                        stack.extend(input[::-1] + [input[mnemonic_num - 1]])
                    case "SWAP":
                        top = input[0]
                        input[0] = input[mnemonic_num]
                        input[mnemonic_num] = top
                        stack.extend(input[::-1])
                    case "LOG":
                        offset = input[0]
                        size = input[1]
                        context.gas -= memory.extend(offset + size)
                        context.gas -= 8 * size
                    case "CREATE":
                        value, offset, size = input
                        contract_address = compute_contract_address(context.address, 0)  # TODO: fix

                        child_context = copy.deepcopy(context)
                        child_context.bytecode = memory.get_as_bytes(offset, size)
                        # child_context.gas = call_gas_cost # TODO
                        child_context.address = contract_address
                        child_context.caller = context.address
                        child_context.callvalue = value
                        # child_context.calldata = memory.get_as_bytes(
                        #     args_offset, args_size
                        # )
                        child_context.calldata = b""
                        child_context.return_data = b""
                        child_context.depth += 1

                        result = disassemble(
                            context=child_context,
                            trace=trace,
                            entrypoint=0,
                            max_steps=max_steps,
                            decode_stack=decode_stack,
                            ignore_stack_underflow=ignore_stack_underflow,
                            silent=silent,
                            hide_pc=hide_pc,
                            show_opcodes=show_opcodes,
                            memory_display=memory_display,
                            invocation_only=invocation_only,
                            return_trace_logs=return_trace_logs,
                        )

                        if result.success:
                            context.gas += child_context.gas
                            context.state = child_context.state
                            # result_return_data = result.return_data[
                            #     :ret_size
                            # ] + b"\x00" * max(0, ret_size - len(result.return_data))
                            # memory.store(ret_offset, result_return_data)
                            # context.return_data = result.return_data
                            context.state.set_code(contract_address, result.return_data)
                        else:
                            context.state.original_storages = child_context.state.original_storages
                            context.state.codes = child_context.state.codes
                        context.steps = child_context.steps

                        stack.push(contract_address if result.success else 0)
                        trace_logs.extend(result.trace_logs)
                    case "CALL" | "STATICCALL" | "DELEGATECALL":
                        if mnemonic == "CALL":
                            (
                                gas,
                                address,
                                callvalue,
                                args_offset,
                                args_size,
                                ret_offset,
                                ret_size,
                            ) = input
                            if context.static and callvalue > 0:
                                raise StaticCallError
                        elif mnemonic == "STATICCALL":
                            callvalue = 0
                            (
                                gas,
                                address,
                                args_offset,
                                args_size,
                                ret_offset,
                                ret_size,
                            ) = input
                        elif mnemonic == "DELEGATECALL":
                            callvalue = context.callvalue
                            (
                                gas,
                                address,
                                args_offset,
                                args_size,
                                ret_offset,
                                ret_size,
                            ) = input
                            if context.static:
                                raise StaticCallError

                        memory_cost = memory.extend(ret_offset + ret_size)
                        if callvalue > 0:
                            extra_gas = 2300
                            context.gas -= 9000 - 2300  # TODO
                            # TODO: gas for empty account
                        else:
                            extra_gas = 0
                        if address not in context.state.address_access_set:
                            context.gas -= GAS_CODE_WARM_COLD_DIFF
                            context.state.address_access_set.add(address)
                        code = context.state.get_code(address)
                        call_gas_cost = calculate_message_call_gas(
                            callvalue, gas, context.gas, memory_cost, extra_gas
                        ).cost
                        context.gas -= call_gas_cost
                        child_context = copy.deepcopy(context)
                        child_context.bytecode = code
                        child_context.gas = call_gas_cost
                        child_context.address = address if mnemonic != "DELEGATECALL" else context.address
                        child_context.caller = context.address
                        child_context.callvalue = callvalue
                        child_context.calldata = memory.get_as_bytes(args_offset, args_size)
                        child_context.return_data = b""
                        child_context.depth += 1
                        if mnemonic == "STATICCALL":
                            child_context.static = True

                        result = disassemble(
                            context=child_context,
                            trace=trace,
                            entrypoint=0,
                            max_steps=max_steps,
                            decode_stack=decode_stack,
                            ignore_stack_underflow=ignore_stack_underflow,
                            silent=silent,
                            hide_pc=hide_pc,
                            show_opcodes=show_opcodes,
                            memory_display=memory_display,
                            invocation_only=invocation_only,
                            return_trace_logs=return_trace_logs,
                        )

                        if result.success:
                            context.gas += child_context.gas
                            context.state = child_context.state
                            result_return_data = result.return_data[:ret_size] + b"\x00" * max(
                                0, ret_size - len(result.return_data)
                            )
                            memory.store(ret_offset, result_return_data)
                            context.return_data = result.return_data
                        else:
                            context.state.original_storages = child_context.state.original_storages
                            context.state.codes = child_context.state.codes
                        context.steps = child_context.steps

                        stack.push(int(result.success))
                        trace_logs.extend(result.trace_logs)
                    case "CALLCODE":
                        assert False, "CALLCODE is not supported"
                    case "RETURN":
                        offset, size = input
                        if not silent:
                            instruction_message += (
                                f"\n{'return'.rjust(TAB_SIZE * 2)}{' ' * TAB_SIZE}{memory.get_as_hex(offset, size)}"
                            )
                        return_data = memory.get_as_bytes(offset, size)
                        break_flag = True
                    case "CREATE2":
                        assert False, "CREATE2 is not supported"
                    case "REVERT":
                        offset, size = input
                        return_data = memory.get_as_bytes(offset, size)
                        return_data_hex = memory.get_as_hex(offset, size)
                        if not silent:
                            instruction_message += f"\n{'revert'.rjust(TAB_SIZE * 2)}{' ' * TAB_SIZE}{return_data_hex}"
                            if return_data_hex.startswith("08c379a0"):
                                # Error(strings)
                                arg_data = return_data[0x04:]
                                strings_offset = int(arg_data[:0x20].hex(), 16)
                                strings_size = int(
                                    arg_data[strings_offset : strings_offset + 0x20].hex(),
                                    16,
                                )
                                strings = arg_data[strings_offset + 0x20 : strings_offset + 0x20 + strings_size]
                                instruction_message += f"\n{' ' * (TAB_SIZE * 3)}Error(string): {strings.decode()}"
                        break_flag = True
                        success = False
                    case "INVALID":
                        break_flag = True
                        success = False
                    case "SELFDESTRUCT":
                        assert False, "SELFDESTRUCT is not supported"
                        break_flag = True
                    case _:
                        assert False, "Invalid opcode"
                        break_flag = True
                        success = False

            except StaticCallError as e:
                print("Static call revert", e, file=sys.stderr)
                break_flag = True
                success = False
            except Exception as e:
                print("Error", e, file=sys.stderr)
                break_flag = True
                success = False

        if trace and not silent:
            instruction_message += f"\n{'stack'.rjust(TAB_SIZE * 2)}{' ' * TAB_SIZE}{stack.to_string()}"
            if decode_stack:
                instruction_message += f"\n{' ' * (TAB_SIZE * 3)}{stack.to_string_with_decode()}"

            if memory_display == MemoryDisplay.ALWAYS or (
                memory_display == MemoryDisplay.ONCHANGE
                and mnemonic
                in [
                    "MLOAD",
                    "MSTORE",
                    "MSTORE8",
                    "MCOPY",
                    "CALL",
                    "STATICCALL",
                    "CALLCODE",
                    "DELEGATECALL",
                    "CODECOPY",
                    "EXTCODECOPY",
                    "RETURNDATACOPY",
                    "CALLDATACOPY",
                    "LOG",
                    "CREATE",
                    "CREATE2",
                    "RETURN",
                    "REVERT",
                ]
            ):
                lines = memory.to_string(memory_range=memory_range)
                for i, line in enumerate(lines):
                    if i == 0:
                        instruction_message += f"\n{'memory'.rjust(TAB_SIZE * 2)}{' ' * TAB_SIZE}"
                    else:
                        instruction_message += f"\n{' ' * (TAB_SIZE * 3)}"
                    instruction_message += f"{line}"

        if not silent:
            if invocation_only:
                if is_invocation_mnemonic(mnemonic):
                    print(instruction_message)
            else:
                print(instruction_message)

        context.steps += 1
        if context.steps >= max_steps:
            break
        if break_flag:
            break
        pc = next_pc

    if not silent:
        print()
        if warning_messages != "":
            print(Colors.YELLOW + "WARNING:")
            print(warning_messages + Colors.ENDC)

    disassemble_result = DisassembleResult(
        last_jump_to_address,
        disassembled_code,
        trace_logs if trace and return_trace_logs else [],
        success,
        return_data,
        stack,
        memory,
    )

    return disassemble_result



from .colors import Colors
from .utils import pad_even

NodeValue = str | int | list["Node"]


class Node:
    type: str
    value: NodeValue

    mnemonic_num: int | None
    input_count: int | None

    def __init__(
        self,
        type_: str,
        value: NodeValue,
        mnemonic_num: int | None = None,
        input_count: int | None = None,
    ) -> None:
        self.type = type_
        self.value = value

        self.mnemonic_num = mnemonic_num
        self.input_count = input_count

    @staticmethod
    def unwrap(s: "str | int | Node") -> str:
        s = str(s)
        return s[1:-1] if s[0] == "(" and s[-1] == ")" else s

    def evaluate(self) -> str:
        if self.type != "uint256" and self.type != "var":
            if self.type == "EQ":
                assert isinstance(self.value, list)
                if self.value[0].type == "uint256" and self.value[1].type == "uint256":
                    return str(self.value[0].value == self.value[1].value).lower()
        return ""

    def __repr__(self) -> str:
        if self.type == "uint256":
            assert isinstance(self.value, int), f"{self.type} {type(self.value)} {self.value}"
            return f"{pad_even(hex(self.value))}"
        elif self.type == "var":
            assert isinstance(self.value, str), f"{self.type} {type(self.value)} {self.value}"
            return self.value
        else:
            assert isinstance(self.value, list), f"{self.type} {type(self.value)} {self.value}"
            match self.type:
                # case "STOP":
                case "ADD":
                    return f"({self.value[0]} + {self.value[1]})"
                case "MUL":
                    return f"({self.value[0]} * {self.value[1]})"
                case "SUB":
                    return f"({self.value[0]} - {self.value[1]})"
                case "DIV":
                    return f"({self.value[0]} / {self.value[1]})"
                case "SDIV":
                    return f"(int256({Node.unwrap(self.value[0])}) / int256({Node.unwrap(self.value[1])}))"
                case "MOD":
                    return f"({self.value[0]} % {self.value[1]})"
                case "SMOD":
                    return f"(int256({Node.unwrap(self.value[0])}) % int256({Node.unwrap(self.value[1])}))"
                case "ADDMOD":
                    return f"(({self.value[0]} + {self.value[1]}) % {self.value[2]})"
                case "MULMOD":
                    return f"(({self.value[0]} * {self.value[1]}) % {self.value[2]})"
                case "EXP":
                    return f"({self.value[0]} ** {self.value[1]})"
                # case "SIGNEXTEND":
                case "LT":
                    return f"({self.value[0]} < {self.value[1]})"
                case "GT":
                    return f"({self.value[0]} > {self.value[1]})"
                case "SLT":
                    return f"(int256({Node.unwrap(self.value[0])}) < int256({Node.unwrap(self.value[1])}))"
                case "SGT":
                    return f"(int256({Node.unwrap(self.value[0])}) > int256({Node.unwrap(self.value[1])}))"
                case "EQ":
                    return f"({self.value[0]} == {self.value[1]})"
                # case "ISZERO":
                case "AND":
                    return f"({self.value[0]} & {self.value[1]})"
                case "OR":
                    return f"({self.value[0]} | {self.value[1]})"
                case "XOR":
                    return f"({self.value[0]} ^ {self.value[1]})"
                # case "NOT":
                # case "BYTE":
                case "SHL":
                    return f"({self.value[1]} << {self.value[0]})"
                case "SHR":
                    return f"({self.value[1]} >> {self.value[0]})"
                case "SAR":
                    return f"(int256({Node.unwrap(self.value[1])}) >> {self.value[0]})"
                # case "KECCAK256":
                # case "ADDRESS":
                # case "BALANCE":
                # case "ORIGIN":
                # case "CALLER":
                # case "CALLVALUE":
                # case "CALLDATALOAD":
                # case "CALLDATASIZE":
                # case "CALLDATACOPY":
                # case "CODESIZE":
                # case "CODECOPY":
                # case "GASPRICE":
                # case "EXTCODESIZE":
                # case "EXTCODECOPY":
                # case "RETURNDATASIZE":
                # case "RETURNDATACOPY":
                # case "EXTCODEHASH":
                # case "BLOCKHASH":
                # case "COINBASE":
                # case "TIMESTAMP":
                # case "NUMBER":
                # case "PREVRANDAO":
                # case "GASLIMIT":
                # case "CHAINID":
                # case "SELFBALANCE":
                # case "BASEFEE":
                case "POP":
                    return f"{Colors.GRAY}{Colors.BOLD}{self.type}{Colors.ENDC}{Colors.GRAY}(){Colors.ENDC} # {Node.unwrap(self.value[0])}"
                # case "MLOAD":
                # case "MSTORE":
                # case "MSTORE8":
                # case "SLOAD":
                # case "SSTORE":
                case "JUMP":
                    return f"{Colors.CYAN}{Colors.BOLD}{self.type}{Colors.ENDC}({Node.unwrap(self.value[0])})"
                case "JUMPI":
                    return f"{Colors.CYAN}{Colors.BOLD}{self.type}{Colors.ENDC}({Node.unwrap(self.value[0])}, {Node.unwrap(self.value[1])})"
                # case "PC":
                # case "MSIZE":
                # case "GAS":
                case "JUMPDEST":
                    return f"{Colors.BLUE}{Colors.BOLD}{self.type}{Colors.ENDC}{Colors.BLUE}(){Colors.ENDC}"
                case "PUSH":
                    if self.mnemonic_num == 0:
                        return f"{Colors.BOLD}{self.type}{self.mnemonic_num}{Colors.ENDC}"
                    else:
                        return f"{Colors.BOLD}{self.type}{self.mnemonic_num}{Colors.ENDC}({self.value[0]})"
                case "DUP":
                    ret = f"{Colors.BOLD}{self.type}{self.mnemonic_num}{Colors.ENDC}() # "
                    assert isinstance(self.mnemonic_num, int)
                    if self.mnemonic_num >= 2:
                        ret += "..., "
                    ret += f"{str(self.value[-1])}"
                    return ret
                case "SWAP":
                    ret = f"{Colors.BOLD}{self.type}{self.mnemonic_num}{Colors.ENDC}() # "
                    assert isinstance(self.mnemonic_num, int)
                    if self.mnemonic_num >= 2:
                        ret += f"{str(self.value[0])}, ..., {str(self.value[-1])}"
                    else:
                        ret += f"{str(self.value[0])}, {str(self.value[-1])}"
                    return ret
                case "LOG":
                    return f"{Colors.BOLD}{self.type}{self.mnemonic_num}{Colors.ENDC}({str(self.value)[1:-1]})"
                # case "CREATE":
                # case "CALL":
                # case "CALLCODE":
                # case "RETURN":
                # case "DELEGATECALL":
                # case "CREATE2":
                # case "STATICCALL":
                # case "REVERT":
                # case "INVALID":
                # case "SELFDESTRUCT":
                case _:
                    if self.input_count == 1:
                        return f"{Colors.BOLD}{self.type}{Colors.ENDC}({Node.unwrap(self.value[0])})"
                    else:
                        return f"{Colors.BOLD}{self.type}{Colors.ENDC}({str(self.value)[1:-1]})"



from erever.secp256k1 import ECDSA, ECPoint, Fp, G, n


def test_key_pair() -> None:
    # openssl ecparam -genkey -name secp256k1 -text -noout -outform DER | xxd -p -c 1000 | sed 's/41534e31204f49443a20736563703235366b310a30740201010420/\nPrivate Key: /' | sed 's/a00706052b8104000aa144034200/\'$'\nPublic Key: /'
    private_key = 0xFAE0CE9EE78E6C3BD4AC3CEC2C18932AE1D374DBA0EF7A4D76024D1A49286C60
    public_key = ECPoint(
        Fp(0xF6A49909F37B380914450718E75D53443F3C5848DF95A2DFE684308B700859A5),
        Fp(0x5C5556D2225A89747AB2BF12808B24D7A663282CA7A32849C0D47BB9710AE979),
    )
    assert private_key * G == public_key

    m = 0x1234
    r, s = ECDSA.sign(m, private_key)
    assert ECDSA.verify(public_key, m, r, s)


def test_half_g_fun_fact() -> None:
    half_G = G * pow(2, -1, n)
    assert int(half_G.x) == 0x3B7_8CE563F89A0ED9414F5AA28AD0D96D6795F9C6_3  # 166 bits



OPCODES: dict[int, tuple[str, int, int, int, str, tuple[str, ...]]] = {
    # value: (mnemonic, stack input count, stack output count, base_gas, description, stack input names)
    0x00: ("STOP", 0, 0, 0, "Halts execution.", ()),
    0x01: ("ADD", 2, 1, 3, "Addition operation.", ("a", "b")),
    0x02: ("MUL", 2, 1, 5, "Multiplication operation.", ("a", "b")),
    0x03: ("SUB", 2, 1, 3, "Subtraction operation.", ("a", "b")),
    0x04: ("DIV", 2, 1, 5, "Integer division operation.", ("a", "b")),
    0x05: (
        "SDIV",
        2,
        1,
        5,
        "Signed integer division operation (truncated).",
        ("a", "b"),
    ),
    0x06: ("MOD", 2, 1, 5, "Modulo remainder operation.", ("a", "b")),
    0x07: ("SMOD", 2, 1, 5, "Signed modulo remainder operation.", ("a", "b")),
    0x08: ("ADDMOD", 3, 1, 8, "Modulo addition operation.", ("a", "b", "mod")),
    0x09: ("MULMOD", 3, 1, 8, "Modulo multiplication operation.", ("a", "b", "mod")),
    0x0A: ("EXP", 2, 1, 10, "Exponential operation.", ("x", "exp")),
    0x0B: (
        "SIGNEXTEND",
        2,
        1,
        5,
        "Extend length of two's complement signed integer.",
        ("size", "x"),
    ),
    0x10: ("LT", 2, 1, 3, "Less-than comparison.", ("a", "b")),
    0x11: ("GT", 2, 1, 3, "Greater-than comparison.", ("a", "b")),
    0x12: ("SLT", 2, 1, 3, "Signed less-than comparison.", ("a", "b")),
    0x13: ("SGT", 2, 1, 3, "Signed greater-than comparison.", ("a", "b")),
    0x14: ("EQ", 2, 1, 3, "Equality comparison.", ("a", "b")),
    0x15: ("ISZERO", 1, 1, 3, "Simple not operator.", ("x",)),
    0x16: ("AND", 2, 1, 3, "Bitwise AND operation.", ("a", "b")),
    0x17: ("OR", 2, 1, 3, "Bitwise OR operation.", ("a", "b")),
    0x18: ("XOR", 2, 1, 3, "Bitwise XOR operation.", ("a", "b")),
    0x19: ("NOT", 1, 1, 3, "Bitwise NOT operation.", ("x",)),
    0x1A: ("BYTE", 2, 1, 3, "Retrieve single byte from word.", ("i", "x")),
    0x1B: ("SHL", 2, 1, 3, "Left shift operation.", ("shift", "x")),
    0x1C: ("SHR", 2, 1, 3, "Logical right shift operation.", ("shift", "x")),
    0x1D: (
        "SAR",
        2,
        1,
        3,
        "Arithmetic (signed) right shift operation.",
        ("shift", "x"),
    ),
    0x20: ("KECCAK256", 2, 1, 30, "Compute Keccak-256 hash.", ("offset", "size")),
    0x30: ("ADDRESS", 0, 1, 2, "Get address of currently executing account.", ()),
    0x31: ("BALANCE", 1, 1, 100, "Get balance of the given account.", ("addr",)),
    0x32: ("ORIGIN", 0, 1, 2, "Get execution origination address.", ()),
    0x33: ("CALLER", 0, 1, 2, "Get caller address.", ()),
    0x34: (
        "CALLVALUE",
        0,
        1,
        2,
        "Get deposited value by the instruction/transaction responsible for this execution.",
        (),
    ),
    0x35: (
        "CALLDATALOAD",
        1,
        1,
        3,
        "Get input data of current environment.",
        ("offset",),
    ),
    0x36: (
        "CALLDATASIZE",
        0,
        1,
        2,
        "Get size of input data in current environment.",
        (),
    ),
    0x37: (
        "CALLDATACOPY",
        3,
        0,
        3,
        "Copy input data in current environment to memory.",
        ("memOffset", "offset", "size"),
    ),
    0x38: ("CODESIZE", 0, 1, 2, "Get size of code running in current environment.", ()),
    0x39: (
        "CODECOPY",
        3,
        0,
        3,
        "Copy code running in current environment to memory.",
        ("memOffset", "offset", "size"),
    ),
    0x3A: ("GASPRICE", 0, 1, 2, "Get price of gas in current environment.", ()),
    0x3B: ("EXTCODESIZE", 1, 1, 100, "Get size of an account's code.", ("addr",)),
    0x3C: (
        "EXTCODECOPY",
        4,
        0,
        100,
        "Copy an account's code to memory.",
        ("addr", "memOffset", "offset", "size"),
    ),
    0x3D: (
        "RETURNDATASIZE",
        0,
        1,
        2,
        "Get size of output data from the previous call from the current environment.",
        (),
    ),
    0x3E: (
        "RETURNDATACOPY",
        3,
        0,
        3,
        "Copy output data from the previous call to memory.",
        ("memOffset", "offset", "size"),
    ),
    0x3F: ("EXTCODEHASH", 1, 1, 100, "Get hash of an account's code.", ("addr",)),
    0x40: (
        "BLOCKHASH",
        1,
        1,
        20,
        "Get the hash of one of the 256 most recent complete blocks.",
        ("number",),
    ),
    0x41: ("COINBASE", 0, 1, 2, "Get the current block's beneficiary address.", ()),
    0x42: ("TIMESTAMP", 0, 1, 2, "Get the current block's timestamp.", ()),
    0x43: ("NUMBER", 0, 1, 2, "Get the current block's number.", ()),
    0x44: ("PREVRANDAO", 0, 1, 2, "Get the previous block's RANDAO mix.", ()),
    0x45: ("GASLIMIT", 0, 1, 2, "Get the current block's gas limit.", ()),
    0x46: ("CHAINID", 0, 1, 2, "Get the chain ID.", ()),
    0x47: ("SELFBALANCE", 0, 1, 5, "Get balance of currently executing account.", ()),
    0x48: ("BASEFEE", 0, 1, 2, "Get the base fee.", ()),
    0x50: ("POP", 1, 0, 2, "Remove item from stack.", ("x",)),
    0x51: ("MLOAD", 1, 1, 3, "Load word from memory.", ("offset",)),
    0x52: ("MSTORE", 2, 0, 3, "Save word to memory.", ("offset", "x")),
    0x53: ("MSTORE8", 2, 0, 3, "Save byte to memory.", ("offset", "x")),
    0x54: ("SLOAD", 1, 1, 100, "Load word from storage.", ("key",)),
    0x55: ("SSTORE", 2, 0, 100, "Save word to storage.", ("key", "x")),
    0x56: ("JUMP", 1, 0, 8, "Alter the program counter.", ("pc",)),
    0x57: (
        "JUMPI",
        2,
        0,
        10,
        "Conditionally alter the program counter.",
        ("pc", "cond"),
    ),
    0x58: (
        "PC",
        0,
        1,
        2,
        "Get the value of the program counter prior to the increment corresponding to this instruction.",
        (),
    ),
    0x59: ("MSIZE", 0, 1, 2, "Get the size of active memory in bytes.", ()),
    0x5A: (
        "GAS",
        0,
        1,
        2,
        "Get the amount of available gas, including the corresponding reduction for the cost of this instruction.",
        (),
    ),
    0x5B: ("JUMPDEST", 0, 0, 1, "Mark a valid destination for jumps.", ()),
    0x5E: (
        "MCOPY",
        3,
        0,
        3,
        "Copy word from memory to memory.",
        ("destOffset", "offset", "size"),
    ),
    0x5F: ("PUSH0", 0, 0, 2, "Place 0x00 on stack.", ()),
    0x60: ("PUSH1", 0, 1, 3, "Place 1 byte item on stack.", ()),
    0x61: ("PUSH2", 0, 1, 3, "Place 2 byte item on stack.", ()),
    0x62: ("PUSH3", 0, 1, 3, "Place 3 byte item on stack.", ()),
    0x63: ("PUSH4", 0, 1, 3, "Place 4 byte item on stack.", ()),
    0x64: ("PUSH5", 0, 1, 3, "Place 5 byte item on stack.", ()),
    0x65: ("PUSH6", 0, 1, 3, "Place 6 byte item on stack.", ()),
    0x66: ("PUSH7", 0, 1, 3, "Place 7 byte item on stack.", ()),
    0x67: ("PUSH8", 0, 1, 3, "Place 8 byte item on stack.", ()),
    0x68: ("PUSH9", 0, 1, 3, "Place 9 byte item on stack.", ()),
    0x69: ("PUSH10", 0, 1, 3, "Place 10 byte item on stack.", ()),
    0x6A: ("PUSH11", 0, 1, 3, "Place 11 byte item on stack.", ()),
    0x6B: ("PUSH12", 0, 1, 3, "Place 12 byte item on stack.", ()),
    0x6C: ("PUSH13", 0, 1, 3, "Place 13 byte item on stack.", ()),
    0x6D: ("PUSH14", 0, 1, 3, "Place 14 byte item on stack.", ()),
    0x6E: ("PUSH15", 0, 1, 3, "Place 15 byte item on stack.", ()),
    0x6F: ("PUSH16", 0, 1, 3, "Place 16 byte item on stack.", ()),
    0x70: ("PUSH17", 0, 1, 3, "Place 17 byte item on stack.", ()),
    0x71: ("PUSH18", 0, 1, 3, "Place 18 byte item on stack.", ()),
    0x72: ("PUSH19", 0, 1, 3, "Place 19 byte item on stack.", ()),
    0x73: ("PUSH20", 0, 1, 3, "Place 20 byte item on stack.", ()),
    0x74: ("PUSH21", 0, 1, 3, "Place 21 byte item on stack.", ()),
    0x75: ("PUSH22", 0, 1, 3, "Place 22 byte item on stack.", ()),
    0x76: ("PUSH23", 0, 1, 3, "Place 23 byte item on stack.", ()),
    0x77: ("PUSH24", 0, 1, 3, "Place 24 byte item on stack.", ()),
    0x78: ("PUSH25", 0, 1, 3, "Place 25 byte item on stack.", ()),
    0x79: ("PUSH26", 0, 1, 3, "Place 26 byte item on stack.", ()),
    0x7A: ("PUSH27", 0, 1, 3, "Place 27 byte item on stack.", ()),
    0x7B: ("PUSH28", 0, 1, 3, "Place 28 byte item on stack.", ()),
    0x7C: ("PUSH29", 0, 1, 3, "Place 29 byte item on stack.", ()),
    0x7D: ("PUSH30", 0, 1, 3, "Place 30 byte item on stack.", ()),
    0x7E: ("PUSH31", 0, 1, 3, "Place 31 byte item on stack.", ()),
    0x7F: ("PUSH32", 0, 1, 3, "Place 32-byte (full word) item on stack.", ()),
    0x80: ("DUP1", 1, 2, 3, "Duplicate 1st stack item.", ("x",)),
    0x81: ("DUP2", 2, 3, 3, "Duplicate 2nd stack item.", ("x",)),
    0x82: ("DUP3", 3, 4, 3, "Duplicate 3rd stack item.", ("x",)),
    0x83: ("DUP4", 4, 5, 3, "Duplicate 4th stack item.", ("x",)),
    0x84: ("DUP5", 5, 6, 3, "Duplicate 5th stack item.", ("x",)),
    0x85: ("DUP6", 6, 7, 3, "Duplicate 6th stack item.", ("x",)),
    0x86: ("DUP7", 7, 8, 3, "Duplicate 7th stack item.", ("x",)),
    0x87: ("DUP8", 8, 9, 3, "Duplicate 8th stack item.", ("x",)),
    0x88: ("DUP9", 9, 10, 3, "Duplicate 9th stack item.", ("x",)),
    0x89: ("DUP10", 10, 11, 3, "Duplicate 10th stack item.", ("x",)),
    0x8A: ("DUP11", 11, 12, 3, "Duplicate 11th stack item.", ("x",)),
    0x8B: ("DUP12", 12, 13, 3, "Duplicate 12th stack item.", ("x",)),
    0x8C: ("DUP13", 13, 14, 3, "Duplicate 13th stack item.", ("x",)),
    0x8D: ("DUP14", 14, 15, 3, "Duplicate 14th stack item.", ("x",)),
    0x8E: ("DUP15", 15, 16, 3, "Duplicate 15th stack item.", ("x",)),
    0x8F: ("DUP16", 16, 17, 3, "Duplicate 16th stack item.", ("x",)),
    0x90: ("SWAP1", 2, 2, 3, "Exchange 1st and 2nd stack items.", ("a", "b")),
    0x91: ("SWAP2", 3, 3, 3, "Exchange 1st and 3rd stack items.", ("a", "b")),
    0x92: ("SWAP3", 4, 4, 3, "Exchange 1st and 4th stack items.", ("a", "b")),
    0x93: ("SWAP4", 5, 5, 3, "Exchange 1st and 5th stack items.", ("a", "b")),
    0x94: ("SWAP5", 6, 6, 3, "Exchange 1st and 6th stack items.", ("a", "b")),
    0x95: ("SWAP6", 7, 7, 3, "Exchange 1st and 7th stack items.", ("a", "b")),
    0x96: ("SWAP7", 8, 8, 3, "Exchange 1st and 8th stack items.", ("a", "b")),
    0x97: ("SWAP8", 9, 9, 3, "Exchange 1st and 9th stack items.", ("a", "b")),
    0x98: ("SWAP9", 10, 10, 3, "Exchange 1st and 10th stack items.", ("a", "b")),
    0x99: ("SWAP10", 11, 11, 3, "Exchange 1st and 11th stack items.", ("a", "b")),
    0x9A: ("SWAP11", 12, 12, 3, "Exchange 1st and 12th stack items.", ("a", "b")),
    0x9B: ("SWAP12", 13, 13, 3, "Exchange 1st and 13th stack items.", ("a", "b")),
    0x9C: ("SWAP13", 14, 14, 3, "Exchange 1st and 14th stack items.", ("a", "b")),
    0x9D: ("SWAP14", 15, 15, 3, "Exchange 1st and 15th stack items.", ("a", "b")),
    0x9E: ("SWAP15", 16, 16, 3, "Exchange 1st and 16th stack items.", ("a", "b")),
    0x9F: ("SWAP16", 17, 17, 3, "Exchange 1st and 17th stack items.", ("a", "b")),
    0xA0: ("LOG0", 2, 0, 375, "Append log record with no topics.", ("offset", "size")),
    0xA1: (
        "LOG1",
        3,
        0,
        750,
        "Append log record with one topic.",
        ("offset", "size", "topic1"),
    ),
    0xA2: (
        "LOG2",
        4,
        0,
        1125,
        "Append log record with two topics.",
        ("offset", "size", "topic1", "topic2"),
    ),
    0xA3: (
        "LOG3",
        5,
        0,
        1500,
        "Append log record with three topics.",
        ("offset", "size", "topic1", "topic2", "topic3"),
    ),
    0xA4: (
        "LOG4",
        6,
        0,
        1875,
        "Append log record with four topics.",
        ("offset", "size", "topic1", "topic2", "topic3", "topic4"),
    ),
    0xF0: (
        "CREATE",
        3,
        1,
        32000,
        "Create a new account with associated code.",
        ("value", "offset", "size"),
    ),
    0xF1: (
        "CALL",
        7,
        1,
        100,
        "Message-call into an account.",
        ("gas", "addr", "value", "argsOffset", "argsSize", "retOffset", "retSize"),
    ),
    0xF2: (
        "CALLCODE",
        7,
        1,
        100,
        "Message-call into this account with an alternative account's code.",
        ("gas", "addr", "value", "argsOffset", "argsSize", "retOffset", "retSize"),
    ),
    0xF3: (
        "RETURN",
        2,
        0,
        0,
        "Halt execution returning output data.",
        ("offset", "size"),
    ),
    0xF4: (
        "DELEGATECALL",
        6,
        1,
        100,
        "Message-call into this account with an alternative account's code, but persisting the current values for sender and value.",
        ("gas", "addr", "argsOffset", "argsSize", "retOffset", "retSize"),
    ),
    0xF5: (
        "CREATE2",
        4,
        1,
        32000,
        "Create a new account with associated code at a predictable address",
        ("value", "offset", "size", "salt"),
    ),
    0xFA: (
        "STATICCALL",
        6,
        1,
        100,
        "Static message-call into an account.",
        ("gas", "addr", "argsOffset", "argsSize", "retOffset", "retSize"),
    ),
    0xFD: (
        "REVERT",
        2,
        0,
        0,
        "Halt execution reverting state changes but returning data and remaining gas.",
        ("offset", "size"),
    ),
    0xFE: ("INVALID", 0, 0, 0, "Designated invalid instruction.", ()),  # gas: NaN
    0xFF: (
        "SELFDESTRUCT",
        1,
        0,
        5000,
        "Halt execution and register account for later deletion.",
        ("addr",),
    ),
}



from Crypto.Random.random import randint

# secp256k1

# 2^256 - 2^32 - 2^9 - 2^8 - 2^7 - 2^6 - 2^4 - 1
p = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F

# y^2 = x^3 + 7
a = 0
b = 7


class Fp:
    def __init__(self, x: int) -> None:
        self.x = x % p

    def __add__(self, other: "Fp") -> "Fp":
        return Fp((self.x + other.x) % p)

    def __sub__(self, other: "Fp") -> "Fp":
        return Fp((self.x - other.x) % p)

    def __mul__(self, other: "Fp") -> "Fp":
        return Fp((self.x * other.x) % p)

    def __pow__(self, exponent: int) -> "Fp":
        return Fp(pow(self.x, exponent, p))

    def __truediv__(self, other: "Fp") -> "Fp":
        return self * other ** (p - 2)

    def __neg__(self) -> "Fp":
        return Fp(-self.x)

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, Fp)
        return self.x == other.x

    def __int__(self) -> int:
        return self.x

    def __str__(self) -> str:
        return str(self.x)


class ECPoint:
    def __init__(self, x: Fp, y: Fp) -> None:
        self.x = x
        self.y = y

    def __add__(self, other: "ECPoint") -> "ECPoint":
        if self.is_zero():
            return other
        elif other.is_zero():
            return self
        elif self == other:
            return self._double()
        elif self == -other:
            return ECPoint(Fp(0), Fp(0))
        else:
            return self._add(other)

    def __mul__(self, other: int) -> "ECPoint":
        return self._mul(other)

    def __rmul__(self, other: int) -> "ECPoint":
        return self * other

    def _add(self, other: "ECPoint") -> "ECPoint":
        s = (other.y - self.y) / (other.x - self.x)
        x = s**2 - self.x - other.x
        y = s * (self.x - x) - self.y
        return ECPoint(x, y)

    def _double(self) -> "ECPoint":
        s = (Fp(3) * self.x**2 + Fp(a)) / (Fp(2) * self.y)
        x = s**2 - Fp(2) * self.x
        y = s * (self.x - x) - self.y
        return ECPoint(x, y)

    def _mul(self, other: int) -> "ECPoint":
        if other == 0:
            return ECPoint(Fp(0), Fp(0))
        elif other == 1:
            return self
        elif other % 2 == 0:
            return (self + self) * (other // 2)
        else:
            return self + (self + self) * ((other - 1) // 2)

    def is_zero(self) -> bool:
        return self.x == Fp(0) and self.y == Fp(0)

    def __neg__(self) -> "ECPoint":
        return ECPoint(self.x, -self.y)

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, ECPoint)
        return self.x == other.x and self.y == other.y

    def __str__(self) -> str:
        return f"({self.x.x}, {self.y.x})"


gx = Fp(0x79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798)
gy = Fp(0x483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8)
G = ECPoint(gx, gy)
n = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141


class ECDSA:
    @staticmethod
    def sign(m: int, d: int) -> tuple[int, int]:
        assert m < n
        k = randint(1, n - 1)
        r = int((k * G).x)
        s = (m + r * d) * pow(k, -1, n) % n
        return (r, s)

    @staticmethod
    def verify(Q: ECPoint, m: int, r: int, s: int) -> bool:
        if not (0 < r < n and 0 < s < n):
            return False
        w = pow(s, -1, n)
        u = m * w % n
        v = r * w % n
        return int((u * G + v * Q).x) == r



from .node import Node


class SymbolicStack:
    var_n: int
    stack: list[Node]

    def __init__(self) -> None:
        self.var_n = 0
        self.stack = []

    def push(self, x: Node) -> None:
        self.stack.append(x)

    def extend(self, x: list[Node]) -> None:
        self.stack.extend(x)

    def pop(self) -> Node:
        if len(self.stack) == 0:
            ret = Node("var", f"var_{self.var_n}")
            self.var_n += 1
            return ret
        else:
            return self.stack.pop()

    def __len__(self) -> int:
        return len(self.stack)

    def clear(self) -> None:
        self.stack = []

    def __repr__(self) -> str:
        return repr(self.stack[::-1])

    def to_string(self) -> str:
        return self.__repr__()
